-----------------------------------------------------------------------------------------------
-- Client Lua Script for KiruChallenge
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-- KiruChallenge Module Definition
local KiruChallenge = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants and Properties
-----------------------------------------------------------------------------------------------
local versionNumber = "v0.991"

--todo
local syncsize = 125
local housingZone = 1630

local debugMode = false
local partyOfOne = false

local needRefresh = false

local Status_OK = "Ready"
local Status_Area = "Area"
local Status_Reward = "Reward"
local Status_Cooling = "Cooling"
local Status_NoMore = "NoMore"
local Status_Active = "Active"
local Status_Wait = "Wait"
local Status_Raid = "Raid"
local Status_Housing = "House"

local Status_OK_Color = "FF00FF00" -- green
local Status_Area_Color = "FFFF00FF" -- purple
local Status_Reward_Color = "FFFFFF00" -- yellow
local Status_Cooling_Color = "FF00FFFF" -- cyan
local Status_NoMore_Color = "FF777777" -- gray
local Status_Active_Color = "FFFFFFFF" -- white
local Status_Wait_Color = "FF777777" -- dark gray
local Status_Raid_Color = "FF777777" -- gray
local Status_Housing_Color = "FF777777" --gray

--map icon
local spriteChallenge = "CRB_ChallengeTrackerSprites:sprChallengeTypeGenericLarge"
local spriteLockedChallenge = "CRB_HUDAlerts:sprAlert_Challenge"

--some keys don't match their control character and the special keys
--don't have a printable description so.. table
local tblSpecialKeys = {
	[8] = "BK",
	[9] = "TAB",
	[13] = "ENT",
	[16] = "SHF",
	[17] = "CTR",
	[27] = "ESC",
	[32] = "SPA",
	[33] = "PUP",
	[34] = "PDN",
	[35] = "END",
	[36] = "HOM",
	[37] = "LFT",
	[38] = "UP",
	[39] = "RT",
	[40] = "DWN",
	[45] = "INS",
	[46] = "DEL",
	[106] = "*",
	[107] = "+",
	[109] = "-",
	[110] = ".",
	[111] = "/",
	[112] = "F1",
	[113] = "F2",
	[114] = "F3",
	[115] = "F4",
	[116] = "F5",
	[117] = "F6",
	[118] = "F7",
	[119] = "F8",
	[120] = "F9",
	[121] = "F10",
	[122] = "F11",
	[123] = "F12",
	[144] = "NUM",
	[145] = "SCR",
	[185] = ":",
	[187] = "=",
	[188] = ",",
	[189] = "-",
	[190] = ".",
	[191] = "/",
	[192] = "`",
	[219] = "[",
	[220] = "|",
	[221] = "]"
}

local tbiHousingUnits = {
	-- Small Plug Challenge Trigger Units		
	["Rocket Launcher"]                   = {CID = 989},  --Anti-Air Defense
	["Spirit Zapper"]                     = {CID = 636},  --Bone Pit
	["Cubig Feeder"]                      = {CID = 990},  --Cubig Feeder
	["Ikthian Flying Saucer"]             = {CID = 745},  --Flying Saucer
	["Lightly Wounded Patient"]           = {CID = 1073}, --Medical Station
	["Seriously Wounded Patient"]         = {CID = 1073}, --Medical Station
	["Critically Wounded Patient"]        = {CID = 1073}, --Medical Station
	["Electrostatic Container"]           = {CID = 743},  --Weather Control
	["Air-Infused Crystal"]               = {CID = 631},  --Whirlwind		
		
	-- Large Plug Challenge Trigger Units
	["Research Desk"]                     = {CID = 632},  --Eldan Excavation
	["The Pile!"]                         = {CID = 739},  --Garbage Dump
	["Anomaly Scanner"]                   = {CID = 747},  --Ice Pond
	["Anachronondax"]                     = {CID = 681},  --Large Spider
	["Celebratory Incense"]               = {CID = 740},  --Lopp Party
	["Crazed Fire Elemental"]             = {CID = 988},  --Magma Flow
	["Spigot"]                            = {CID = 649},  --Moonshiner Cabin
	["Book of Elements"]                  = {CID = 971},  --Osun
	["Protostar Hazard Training Console"] = {CID = 974},  --Protostar Hazard
	["Plushie"]                           = {CID = 753},  --Shardspire
	["Call to the Spirits!"]              = {CID = 742},   --Spooky Graveyard
	["Detonator"]              			  = {CID = 963}    --Prospector Plot
}

-----------------------------------------------------------------------------------------------
-- Basic Initialization
-----------------------------------------------------------------------------------------------
function KiruChallenge:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self

	--to store data
	self.Data = {}
	self.Data.Preferences = {}
	self.Data.Preferences.IgnoreList = {}
	self.Data.Preferences.WhiteList = {}
	self.Data.Preferences.AutoStart = {}
	self.PartySyncQueue = {}
	
    return o
end

function KiruChallenge:Init()
	--default Init that Carbine sets up.  Additional initialization in StartUp
	local bHasConfigureFunction = true
	local strConfigureButtonText = "KiruChallenge"
	local tDependencies = { 
		"KiruChalDataLib"
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)

	--to do unit checks in housing, these fire before DocLoaded can. 	
	Apollo.RegisterEventHandler("ChangeWorld", "OnChangeWorld", self)
	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)

end
 
-- KiruChallenge OnLoad
function KiruChallenge:OnLoad()
    -- load the main form, call OnDocLoaded when completed
	self.xmlDoc = XmlDoc.CreateFromFile("KiruChallenge.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-- KiruChallenge OnDocLoaded
function KiruChallenge:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		self.Main = {}		
	    self.Main.Window = Apollo.LoadForm(self.xmlDoc, "KiruChallengeForm", nil, self)
		if self.Main.Window == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
	end
	
	self.DataPackage = Apollo.GetPackage("KiruChalDataLib").tPackage
	
	self.PartySync = {}
	self.Option = {}
	
	self.Main.ItemList = {}	
	self.Main.OutPosition = 0
	
	self.Data.Rewards = {}
	
	--read in from the config data
	if self.Data.Preferences.FromRestore ~= nil then
		local FromRestore = self.Data.Preferences.FromRestore
	
		local RestoreAnchors = FromRestore.MainWindowPos
		if RestoreAnchors then
			self.Main.Window:SetAnchorOffsets(RestoreAnchors.l, RestoreAnchors.t, RestoreAnchors.r, RestoreAnchors.b)
		end	
		
		self.Keybind = FromRestore.Keybind
		self.KeybindNext = FromRestore.KeybindNext
		self.Data.Preferences.ShowKeybindNub = FromRestore.ShowKeybindNub
		self.Data.Preferences.IgnoreList = FromRestore.IgnoreList or {}
		self.Data.Preferences.WhiteList = FromRestore.WhiteList or {}
		self.Data.Preferences.AutoStart = FromRestore.AutoStart or {}
		self.Data.Preferences.HideNoChallenges = FromRestore.HideNoChallenges
		self.Data.Preferences.PartySpam = FromRestore.PartySpam
		self.Data.Preferences.PartySpamLoot = FromRestore.PartySpamLoot
		self.Data.Preferences.PartySync = FromRestore.PartySync
		self.Data.Preferences.ShowMap = FromRestore.ShowMap
		self.Data.Preferences.AutoReward = FromRestore.AutoReward
		self.Data.Preferences.ReadyDistance = FromRestore.ReadyDistance
		self.Data.Preferences.ShowCooldown = FromRestore.ShowCooldown
		self.Data.Preferences.HideCombat = FromRestore.HideCombat
		
		self.Data.Preferences.Opacity = FromRestore.Opacity or 100
		self.Data.Preferences.ShowScroll = FromRestore.ShowScroll or 4
		self.Data.Preferences.FilterDistance = FromRestore.FilterDistance or 2000

		self.Data.Preferences.SyncAutoHide = FromRestore.SyncAutoHide		
		self.Data.Preferences.SyncAutoHideTime = FromRestore.SyncAutoHideTime or 30		
		
		self.PartySync.Position = FromRestore.SyncPosition or "B"
		
		self.Data.Rewards = FromRestore.Rewards
	else
		self.Data.Preferences.IgnoreList = {}
		self.Data.Preferences.WhiteList = {}
		self.Data.Preferences.AutoStart = {}
		self.Data.Preferences.Opacity = 100
		self.Data.Preferences.ShowScroll = 4
		self.Data.Preferences.FilterDistance = 2000
		self.Data.Preferences.SyncAutoHideTime = 30		
		self.PartySync.Position = "B"
	end
	if self.Data.Preferences.ShowCooldown == nil then self.Data.Preferences.ShowCooldown = true end
	if self.Data.Preferences.HideCombat == nil then self.Data.Preferences.HideCombat = false end
	if self.Data.Preferences.FilterDistance == nil then self.Data.Preferences.FilterDistance = 2000 end
	--set up the two parent windows of partysync
	self.PartySync.Dock = self.Main.Window:FindChild("PartySyncDock")
	self.PartySync.Float = Apollo.LoadForm(self.xmlDoc, "PartySyncFloat", nil, self)

	--make sure neither one is shown
	self.PartySync.Dock:Show(false,true)
	self.PartySync.Float:Show(false,true)
	
	if self.Data.Preferences.ShowScroll == nil then self.Data.Preferences.ShowScroll = 4 end
	if self.Data.Preferences.SyncAutoHideTime == nil then self.Data.Preferences.SyncAutoHideTime = 30 end
	if self.PartySync.Position == nil then self.PartySync.Position = "B" end

	--get a reference to the lists
	self.Main.ItemList.Window = self.Main.Window:FindChild("ChallengeItemList")

	--misc UI pieces that might get called a lot
	self.Main.BGTransparency = self.Main.Window:FindChild("TransparentBG")
		
	--get a reference to the main background
	self.Main.Background = {}
	self.Main.Background.Window = self.Main.Window:FindChild("Background")

	--get a reference to the sync stuff
	self.Main.SyncBar = {}
	self.Main.SyncBar.Window = self.Main.Window:FindChild("SyncBar")
	self.Main.SyncBar.Label = self.Main.SyncBar.Window:FindChild("SyncLabel")
	self.Main.SyncBar.Button = self.Main.SyncBar.Window:FindChild("SyncButton")
	--make sure it's hidden
	self:ShowSyncBar(false)
		
	--initialize the listitem storage for window components
	self.Main.ItemList.Items = {}
	
	self:ResizeMainWindow()	
	
	--set opacity
	if self.Data.Preferences.Opacity == nil then self.Data.Preferences.Opacity = 100 end
	local opacityPre = (self.Data.Preferences.Opacity / 100) * 255
	local opacityHex = string.format('%x',opacityPre)
	if opacityPre < 16 then opacityHex = "0"..opacityHex end
	self.Main.BGTransparency:SetBGColor(opacityHex .."FFFFFF")
		
	--initialize the storage
	--set up the chellenge "Types" for readability and active challenge storage
	self.Data.ChallengeTypes = {}
	self.Data.ChallengeTypes[0] = { Name = "Combat", ActiveID = 0, HasActive = false }
	self.Data.ChallengeTypes[1] = { Name = "Ability", ActiveID = 0, HasActive = false }
	self.Data.ChallengeTypes[2] = { Name = "General", ActiveID = 0, HasActive = false }
	self.Data.ChallengeTypes[3] = { Name = "Default", ActiveID = 0, HasActive = false }
	self.Data.ChallengeTypes[4] = { Name = "Item", ActiveID = 0, HasActive = false }
	self.Data.ChallengeTypes[7] = { Name = "Activate", ActiveID = 0, HasActive = false }
	
	--set up the tier names (medals)
	self.Data.Tiers = {}
	self.Data.Tiers[0] = { Name = "None", BGColor = "FF777777", TextColor="FFFFFFFF", Abbrev = ""}
	self.Data.Tiers[1] = { Name = "Bronze", BGColor = "FFCD7F32", TextColor="FF000000", Abbrev = "B" }
	self.Data.Tiers[2] = { Name = "Silver", BGColor = "FFCCCCCC", TextColor="FF000000", Abbrev = "S" }
	self.Data.Tiers[3] = { Name = "Gold", BGColor = "FFFFFF00", TextColor="FF000000", Abbrev = "G" }
	
	self.KeybindShiftDown = 0
	self.Data.AllChallenges = self:PopulateAllChallenges()
	
	--check if we're in housing
	self.IsInHousing = HousingLib.IsHousingWorld()
	--populate the main challenge table from the challenge array
	self:PopulateMainChallengeList()
	-- process / sort the local challenges list
	self:ProcessChallengeList()	

	self.InitSync = false

	--show the main window
	self.Main.Window:Show(true)
		
	-- Register handlers for events, slash commands and timer, etc.
	Apollo.RegisterSlashCommand("kch", "OnKCHSlash", self)
	Apollo.RegisterSlashCommand("kchopt", "OnKCHOptSlash", self)
	Apollo.RegisterSlashCommand("kdebug", "OnKDebugSlash", self)
		
	--add to interface menu and window management panel
	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
	Apollo.RegisterEventHandler("WindowManagementReady", "OnWindowManagementReady", self)
	Apollo.RegisterEventHandler("WindowManagementUpdate", "OnWindowManagementUpdate", self)
	Apollo.RegisterEventHandler("ToggleKiruChallenge", "MenuToggle", self)
	
	--challenge related event stuff
	Apollo.RegisterEventHandler("ChallengeActivate", "OnActivateEvent", self)
	Apollo.RegisterEventHandler("ChallengeShared", "OnShared", self)
	Apollo.RegisterEventHandler("ChallengeFailTime", "OnFailedEvent", self)
	Apollo.RegisterEventHandler("ChallengeFailArea", "OnFailedEvent", self)
	Apollo.RegisterEventHandler("ChallengeFailGeneric", "OnFailedEvent", self)
	Apollo.RegisterEventHandler("ChallengeAbandon", "OnAbandonEvent", self)
	Apollo.RegisterEventHandler("ChallengeCompleted", "OnCompletedEvent", self)
	Apollo.RegisterEventHandler("ChallengeAreaRestriction", "OnAreaRestriction", self)
	Apollo.RegisterEventHandler("ChallengeRewardListReady", "OnRewardListReady", self)
	Apollo.RegisterEventHandler("ChallengeRewardReady", "OnRewardReady", self)
	Apollo.RegisterEventHandler("ChallengeUnlocked", "OnChallengeUnlocked",	self)
	
	-- called when the player changes subzones, used to update the challenge window
	Apollo.RegisterEventHandler("SubZoneChanged", "OnSubZoneChanged", self)
	
	--rudimentary and temporary keybinding until I can add a real one to the keybinding window
	Apollo.RegisterEventHandler("SystemKeyDown", "OnSystemKeyDown", self) 

	-- group events for PartySync
	Apollo.RegisterEventHandler("Group_Join", "OnGroupJoin", self)
	Apollo.RegisterEventHandler("Group_Left", "OnGroupLeft", self)
	Apollo.RegisterEventHandler("Group_Remove", "OnGroupRemove", self)
	Apollo.RegisterEventHandler("Group_MemberPromoted",	"OnLeaderChange", self)	
	
	--zone map icons
	Apollo.RegisterEventHandler("GenericEvent_ZoneMap_ZoneChanged", "OnZoneMapChange", self)
	self.ZoneMap = {}
	self.ZoneMap.Overlay = nil
	self.ZoneMap.Map = nil
	self.ZoneMap.Items = {}

	--combat entered/left spam if Hide in Combat is on	
	self:EnteredCombatSpam(self.Data.Preferences.HideCombat)
	self.InCombat = false
	self.CombatViewState = true
	self.CombatSyncViewState = false

	--start a 1 second timer for refreshes, sync, etc.
	self.Timer = ApolloTimer.Create(1.0, true, "OnTimerEvent", self)
end

function KiruChallenge:OnInterfaceMenuListHasLoaded()
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "KiruChallenge", {"ToggleKiruChallenge", "", ""})
end

function KiruChallenge:OnWindowManagementReady()
    Event_FireGenericEvent("WindowManagementAdd", {wnd = self.Main.Window, strName = "KiruChallenge"})
end

function KiruChallenge:OnWindowManagementUpdate(change)
	if change == nil then return end
	if change.wnd == self.Main.Window then 
		self:ResizeMainWindow()	
    	local bMoveable = change.wnd:IsStyleOn("Moveable")
    	self.Main.Window:SetStyle("Sizable", bMoveable)	
	end
end

function KiruChallenge:OnConfigure()
	self:Button_ShowOption()
end

function KiruChallenge:ShowSyncBar(show)
	self.Main.SyncBar.Window:Show(show,true)
end

function KiruChallenge:UpdateSync(syncState)
	if syncState == "WAIT" then
		self.Main.SyncBar.Label:SetText("Syncing")
		self.Main.SyncBar.Button:Show(false,true)
		return
	end
	--otherwise it's "ON"
	self.Main.SyncBar.Label:SetText("SYNC ON")
	self.Main.SyncBar.Button:Show(true,true)

	local l,t,r,b = self.Main.SyncBar.Label:GetAnchorOffsets()
	if self.PartySync.Position == "H" then
		--hide the button altogether, set the label's right anchor to -5
		self.Main.SyncBar.Label:SetAnchorOffsets(l,t,-5,b)
		self.Main.SyncBar.Button:Show(false,true)		
	else
		--make sure the button is visible, set the label's right anchor to -55
		self.Main.SyncBar.Label:SetAnchorOffsets(l,t,-55,b)
		self.Main.SyncBar.Button:Show(true,true)
		if self.ClosedSyncOnPurpose then
			self.Main.SyncBar.Button:SetText("Show")
			self.Main.SyncBar.Button:ChangeArt("CRB_TechTree:CRB_TechTree_GreenBtn")
		else
			self.Main.SyncBar.Button:SetText("Auto")
			self.Main.SyncBar.Button:ChangeArt("CRB_TechTree:CRB_TechTree_BlueBtn")		
		end
	end	
end

function KiruChallenge:OnChangeWorld()
	--if we were in housing before, and we're changing now,
	--its either to another house or leaving.. either way clear the current house list
	local wasInHouse = self.IsInHousing
	if wasInHouse then self.CurrentHouse = {} end
	self.IsInHousing = HousingLib.IsHousingWorld()
	--if we were in house and still in house (another house) subzone won't fire.
	if wasInHouse and self.IsInHousing then needRefresh = true end
end

function KiruChallenge:OnSubZoneChanged() 
	--doublechecking
	self.IsInHousing = HousingLib.IsHousingWorld()
	self:PopulateMainChallengeList()
end

function KiruChallenge:NotImplemented()
	--placeholder function for development
	if debugMode then self:PrintDebug("Something called NotImplemented") end
end

function KiruChallenge:OnUnitCreated(unit)
	if self.IsInHousing == nil then self.IsInHousing = HousingLib.IsHousingWorld() end
	if not self.IsInHousing then return	end
	if unit == null then return end

	local unitName = unit:GetName()
	if tbiHousingUnits[unitName] ~= nil then
		local challengeID = tbiHousingUnits[unitName].CID				
		if challengeID == 0 then return end
		local unitID = unit:GetId()
		
		if self.CurrentHouse == nil then self.CurrentHouse = {} end
		self.CurrentHouse[challengeID] = {UnitID = unitID, Pos = unit:GetPosition()}
	end
	
	--are we within 10m of that unit?
	--local unit = GameLib.GetUnitById(data.unit)
	--unit:ShowHintArrow()
end

function KiruChallenge:CreateOptionWindow()
	self.Option = {}
    self.Option.Window = Apollo.LoadForm(self.xmlDoc, "KiruChallengeOptionForm", nil, self)
	if self.Option.Window == nil then
		Apollo.AddAddonErrorText(self, "Could not load the option window for some reason.")
		return
	end

	local optionMinHeight = 633
	local optionMinWidth = 435
	--set a minimum size for the options window	
	self.Option.Window:SetSizingMinimum(optionMinWidth ,optionMinHeight)
	
	--set up its tabs
	self.Option.OptionsTab = self.Option.Window:FindChild("TabOptions")
	self.Option.IgnoreTab = self.Option.Window:FindChild("TabIgnoreList")
	self.Option.WhiteListTab = self.Option.Window:FindChild("TabWhiteList")	
	self.Option.AutoStartTab = self.Option.Window:FindChild("TabAutoStart")	
	self.Option.LockedTab = self.Option.Window:FindChild("TabLocked")
	self.Option.OptionsTab:AttachTab(self.Option.IgnoreTab, false)
	self.Option.IgnoreTab:AttachTab(self.Option.WhiteListTab, false)
	self.Option.WhiteListTab:AttachTab(self.Option.AutoStartTab, false)
	self.Option.AutoStartTab:AttachTab(self.Option.LockedTab, false)
	self.Option.OptionsTab:Lock(true)
	self.Option.IgnoreTab:Lock(true)
	self.Option.WhiteListTab:Lock(true)
	self.Option.AutoStartTab:Lock(true)
	self.Option.LockedTab:Lock(true)
	
	self.Option.ItemList = {}
	self.Option.ItemListWhite = {}
	self.Option.ItemList.Items = {}
	self.Option.ItemListWhite.Items = {}
	self.Option.LockedList = {}
	self.Option.LockedList.Items = {}
	self.Option.AutoStartList = {}
	self.Option.AutoStartList.Items = {}

	--get references to the tab item lists
	self.Option.ItemList.Window = self.Option.IgnoreTab:FindChild("ItemList")
	self.Option.ItemListWhite.Window = self.Option.WhiteListTab:FindChild("ItemListWhite")
	self.Option.AutoStartList.Window = self.Option.AutoStartTab:FindChild("ItemListAuto")
	self.Option.LockedList.Window = self.Option.LockedTab:FindChild("ItemListLocked")
	--opacity display label
	self.Option.OpacityLabel = self.Option.Window:FindChild("OpacityGroup"):FindChild("OpacityLabel")

	--radio buttons
	self.Option.RadioSyncTop = self.Option.Window:FindChild("Radio_SyncTop")
	self.Option.RadioSyncTop:SetData({Pos = "T"})
	self.Option.RadioSyncBottom = self.Option.Window:FindChild("Radio_SyncBottom")
	self.Option.RadioSyncBottom:SetData({Pos = "B"})
	self.Option.RadioSyncHide = self.Option.Window:FindChild("Radio_SyncHide")
	self.Option.RadioSyncHide:SetData({Pos = "H"})
	self.Option.RadioSyncFloat = self.Option.Window:FindChild("Radio_SyncFloat")
	self.Option.RadioSyncFloat:SetData({Pos = "F"})
	
	self.Option.Window:FindChild("Radio_SharedPrompt"):SetData(1)
	self.Option.Window:FindChild("Radio_SharedAccept"):SetData(0)
	self.Option.Window:FindChild("Radio_SharedReject"):SetData(2)
	if g_InterfaceOptions.Carbine.eShareChallengePreference == 0 then self.Option.Window:FindChild("Radio_SharedAccept"):SetCheck(true) end
	if g_InterfaceOptions.Carbine.eShareChallengePreference == 1 then self.Option.Window:FindChild("Radio_SharedPrompt"):SetCheck(true) end
	if g_InterfaceOptions.Carbine.eShareChallengePreference == 2 then self.Option.Window:FindChild("Radio_SharedReject"):SetCheck(true) end
	
	--set the version number in title
	self.Option.Window:FindChild("Title"):SetText("KiruChallenge Options - " .. versionNumber)

	--read in from the config data
	if self.Data.Preferences.FromRestore ~= nil then
		local FromRestore = self.Data.Preferences.FromRestore
		RestoreAnchors = FromRestore.OptionWindowPos
		if RestoreAnchors then
			--make sure the size meets the minimum
			if (RestoreAnchors.b - RestoreAnchors.t) < optionMinHeight then RestoreAnchors.b = RestoreAnchors.t + optionMinHeight end
			if (RestoreAnchors.r - RestoreAnchors.l) < optionMinWidth then RestoreAnchors.r = RestoreAnchors.l + optionMinWidth end
			self.Option.Window:SetAnchorOffsets(RestoreAnchors.l, RestoreAnchors.t, RestoreAnchors.r, RestoreAnchors.b)
		end	
		if self.Keybind ~= nil then 
			local textbox = self.Option.Window:FindChild("KeybindEditbox")
			local keytext = tblSpecialKeys[self.Keybind]
			if keytext == nil then keytext = string.char(self.Keybind) end
			textbox:SetText(keytext)
		end
		if self.KeybindNext ~= nil then 
			local textbox = self.Option.Window:FindChild("KeybindNextEditbox")
			local keytext = tblSpecialKeys[self.KeybindNext]
			if keytext == nil then keytext = string.char(self.KeybindNext) end
			textbox:SetText(keytext)
		end
		local checkbox = self.Option.Window:FindChild("Button_Hide_No")
		checkbox:SetCheck(self.Data.Preferences.HideNoChallenges)
		checkbox = self.Option.Window:FindChild("Button_PartySpam")
		checkbox:SetCheck(self.Data.Preferences.PartySpam)
		checkbox = self.Option.Window:FindChild("Button_PartySpamLoot")
		checkbox:SetCheck(self.Data.Preferences.PartySpamLoot)
		checkbox = self.Option.Window:FindChild("Button_PartySync")
		checkbox:SetCheck(self.Data.Preferences.PartySync)
		checkbox = self.Option.Window:FindChild("Button_AutoHide")
		checkbox:SetCheck(self.Data.Preferences.SyncAutoHide)
		checkbox = self.Option.Window:FindChild("Button_ShowMap")
		checkbox:SetCheck(self.Data.Preferences.ShowMap)
		checkbox = self.Option.Window:FindChild("Button_ShowNub")
		checkbox:SetCheck(self.Data.Preferences.ShowKeybindNub)
		checkbox = self.Option.Window:FindChild("Button_AutoReward")
		checkbox:SetCheck(self.Data.Preferences.AutoReward)
		checkbox = self.Option.Window:FindChild("Button_ReadyDistance")
		checkbox:SetCheck(self.Data.Preferences.ReadyDistance)
		checkbox = self.Option.Window:FindChild("Button_ShowCooldown")
		checkbox:SetCheck(self.Data.Preferences.ShowCooldown)
		checkbox = self.Option.Window:FindChild("Button_HideCombat")
		checkbox:SetCheck(self.Data.Preferences.HideCombat)
	end	
	--enable or disable the additional spam option
	if not self.Data.Preferences.PartySpam then
		self.Option.Window:FindChild("Button_PartySpamLoot"):Enable(false)
	end
	
	--opacity
	local slider = self.Option.Window:FindChild("Slider_Opacity")
	local slidervalue = (self.Data.Preferences.Opacity / 100) * 255
	slider:SetValue(slidervalue)
	self.Option.OpacityLabel:SetText(self.Data.Preferences.Opacity.."%")
	
	slider = self.Option.Window:FindChild("Slider_Distance")
	slider:SetValue(self.Data.Preferences.FilterDistance)
	if self.Data.Preferences.FilterDistance == 2000 then
		self.Option.Window:FindChild("DistanceLabel"):SetText("Show All")
	else
		self.Option.Window:FindChild("DistanceLabel"):SetText(self.Data.Preferences.FilterDistance.."m")
	end
		--position of partysync
	if self.PartySync.Position == "T" then
		self.Option.RadioSyncTop:SetCheck(true)
	elseif self.PartySync.Position == "H" then
		self.ClosedSyncOnPurpose = true
		self.Option.RadioSyncHide:SetCheck(true)	
	elseif self.PartySync.Position == "F" then
		self.Option.RadioSyncFloat:SetCheck(true)	
	else
		self.Option.RadioSyncBottom:SetCheck(true)	
	end
	--if sync isnt on, disable the radios
	if not self.Data.Preferences.PartySync then
		self.Option.RadioSyncBottom:Enable(false)
		self.Option.RadioSyncHide:Enable(false)
		self.Option.RadioSyncTop:Enable(false)
		self.Option.RadioSyncFloat:Enable(false)
	end
	
	--display scroll settings
	self.Option.Window:FindChild("ShowScrollEdit"):SetText(self.Data.Preferences.ShowScroll)

	--autohide settings
	self.Option.Window:FindChild("AutoHideEdit"):SetText(self.Data.Preferences.SyncAutoHideTime)
	
	--finally, load the lists and populate the statistics
	self:RepopulateOptionList()

	self.Option.Window:Show(true, true)	
end

function KiruChallenge:RepopulateOptionList(reloadCompleted)
	--load the ignore/whitelists
	local totalCompleted = self:PopulateOptionList(reloadCompleted)
	if totalCompleted == nil then totalCompleted = 0 end
	local locked, total = self.DataPackage:GetCount()
	--load the locked list
	self:PopulateOptionLockedList()
	
	--and update the challenge unlocked and completed label
	self.Option.Window:FindChild("ChallengeCount"):SetText("# Challenges Unlocked:  " .. self:GetTableLength(self.Data.AllChallenges) .. "/".. total)
	self.Option.Window:FindChild("ChallengeCountComp"):SetText("Completed:  " .. totalCompleted)
end

-----------------------------------------------------------------------------------------------
-- Settings Persistance
-----------------------------------------------------------------------------------------------

function KiruChallenge:OnSave(eType)
	--character-level saving
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	
	local preferences = self.Data.Preferences
	
	local saveData = {}
	--save main window position/size
	local l,t,r,b = self.Main.Window:GetAnchorOffsets()
	
	--if sync is open and sync is at the top, need to shift the main window position down
	if self.PartySync.IsShown and self.PartySync.Position == "T" then
		t = t + syncsize
		b = b + syncsize
	end
	
	saveData.MainWindowPos = { l = l, t = t, r = r, b = b}
	
	--if the sync is floating, save its size
	if self.PartySync.Position == "F" then
		l,t,r,b = self.PartySync.Float:GetAnchorOffsets()
		saveData.PartySyncWindowPos = { l = l, t = t, r = r, b = b}
	else
		--otherwise persist its last known location
		saveData.PartySyncWindowPos = preferences.FromRestore.PartySyncWindowPos				
	end
	
	--if the option window is open at time of saving, save its current position
	if self.Option ~= nil and self.Option.Window ~= nil then
		l,t,r,b = self.Option.Window:GetAnchorOffsets()
		saveData.OptionWindowPos = { l = l, t = t, r = r, b = b}
	elseif preferences.FromRestore ~= nil then
		--otherwise persist its last known location
		saveData.OptionWindowPos = preferences.FromRestore.OptionWindowPos			
	end
	
	saveData.IgnoreList = preferences.IgnoreList
	saveData.WhiteList = preferences.WhiteList
	saveData.AutoStart = preferences.AutoStart
	saveData.PartySpam = preferences.PartySpam
	saveData.PartySpamLoot = preferences.PartySpamLoot
	saveData.PartySync = preferences.PartySync
	saveData.Keybind = self.Keybind
	saveData.KeybindNext = self.KeybindNext
	saveData.Opacity = preferences.Opacity
	saveData.HideNoChallenges = preferences.HideNoChallenges
	saveData.ShowScroll = preferences.ShowScroll
	saveData.SyncPosition = self.PartySync.Position
	saveData.SyncAutoHide = preferences.SyncAutoHide
	saveData.SyncAutoHideTime = preferences.SyncAutoHideTime
	saveData.ShowMap = preferences.ShowMap
	saveData.ShowKeybindNub = preferences.ShowKeybindNub
	saveData.AutoReward = preferences.AutoReward
	saveData.ReadyDistance = preferences.ReadyDistance
	saveData.FilterDistance = preferences.FilterDistance
	saveData.ShowCooldown = preferences.ShowCooldown
	saveData.HideCombat = preferences.HideCombat
	
	return saveData
end

function KiruChallenge:OnRestore(eType, tSavedData)
	--character-level saving
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	self.Data.Preferences.FromRestore = tSavedData	
end

-----------------------------------------------------------------------------------------------
-- Slash Command, Menu and Challenge Event Calls
-----------------------------------------------------------------------------------------------
-- /kch
function KiruChallenge:OnKCHSlash(inputCmd, inputArgs)

	if inputArgs == "hide" then
		self.Main.Window:Show(false)
		self.CombatViewState = false
		self.ClosedOnPurpose = true
		return
	end
	
	if inputArgs == "poo" then
		self:OnKCHPOOSlash()
		return
	end
	
	if inputArgs == "test" then
		self:OnKCHTestSlash()
		return
	end
	
	if inputArgs == "spam" then
		self:OnKCHSpamSlash()
		return
	end

	if inputArgs == "spam on" then
		self:OnKCHSpamSlash(true)
		return
	end
	
	if inputArgs == "spam off" then
		self:OnKCHSpamSlash(false)
		return
	end
	
	if inputArgs == "?" or inputArgs == "help" then
		Print("KiruChallenge Slash Commands:")
		Print(" /kch           Shows KiruChallenge if it is hidden.")
		Print(" /kch hide           Hides KiruChallenge.")
		Print(" /kch spam on           Turns Party Spam On.")
		Print(" /kch spam off           Turns Party Spam Off.")
		Print(" /kch help           This message.")
		return
	end

	--no arguments.. default to show
	self.ClosedOnPurpose = false
	self.Main.Window:Show(true)
	self.DelayAutoHide = 10
	self.CombatViewState = true	
end

-- /kchopt
function KiruChallenge:OnKCHOptSlash(inputCmd, inputArgs)
	self:Button_ShowOption()
end

function KiruChallenge:SetFormBGSprite(color)
	self.Main.Background.Window:SetSprite("BK3:UI_BK3_ItemQuality" .. color)
end

-- /kdebug
--turns on or off DebugMode for additional spam
function KiruChallenge:OnKDebugSlash(command,args)
	if args == "" then 
		debugMode = not debugMode 
		if debugMode then Print("KiruChallenge: Debug is now ON") else Print("KiruChallenge: Debug is now OFF") end
		return
	end
	local CID = tonumber(args)
	if CID then self:PrintChallengeData(CID) end
end

-- /kchpoo
-- turns on or off PartyOfOne to receive sync window updates for yourself while solo.
function KiruChallenge:OnKCHPOOSlash()
	partyOfOne = not partyOfOne
	if partyOfOne then 
		Print("KiruChallenge: Party of One is now ON")
		self:ShowSyncBar(true)
		self:UpdateSync("ON")
	else 
		Print("KiruChallenge: Party of One now OFF")
		if not self.HasPartySync then
			self:ShowSyncBar(false)	
		end
	end		
end

--/kchspam
--turns on or off party spam
function KiruChallenge:OnKCHSpamSlash(enable)
	if enable == nil then
		self.Data.Preferences.PartySpam = not self.Data.Preferences.PartySpam
	else
		self.Data.Preferences.PartySpam = enable
	end
	if self.Data.Preferences.PartySpam then
		Print("KiruChallenge: Spam ON. Will send Start and Finish messages to Party.")
		if self.Data.Preferences.PartySpamLoot then Print("KiruChallenge: Loot messages too!") end
		--enable loot spam
		if self.Option and self.Option.Window then self.Option.Window:FindChild("Button_PartySpamLoot"):Enable(true) end
	else
		Print("KiruChallenge: Spam OFF. No messages will be sent to Party.")
		--disable loot spam
	end
	if self.Option and self.Option.Window then 
		self.Option.Window:FindChild("Button_PartySpam"):SetCheck(self.Data.Preferences.PartySpam) 
		self.Option.Window:FindChild("Button_PartySpamLoot"):Enable(self.Data.Preferences.PartySpam) 
	end
end

-- /kchtest
-- sends a test ping over the addon channel.
function KiruChallenge:OnKCHTestSlash()
	Print("Preparing to Send a test message")
	if self.ChatChannel == nil then 
		Print("No channel to send test message to.") 
	else
		local msg = { Text = "PING" }
		local test = self.ChatChannel:SendMessage(msg)
		if test then Print("Test sent.") else Print("Test failed to send.") end	
	end
	
	Print("Sending a Test Message to Sync Window")	
	local syncMsg = {}
	syncMsg.Text = "FINAL"
	syncMsg.Player = self.PlayerName
	syncMsg.CID = 99999
	syncMsg.CName = "Test Challenge"
	syncMsg.Status = 1
	syncMsg.Tier = 1
	syncMsg.DTier = 1
	syncMsg.IsTime = false
	syncMsg.Time = "0:00"
	syncMsg.HasGold = true
	syncMsg.Current = "100%"				
	--send it to ourself
	self:OnChannelChat("self", syncMsg)
	
	--send another
	syncMsg.Text = "FINAL"
	syncMsg.Player = "Test User 2"
	syncMsg.CID = 99999
	syncMsg.CName = "Test Challenge"
	syncMsg.Status = 0
	syncMsg.Tier = 2
	syncMsg.DTier = 2
	syncMsg.IsTime = false
	syncMsg.Time = "0:00"
	--some challenges don't have medals, just a single tier of completion
	syncMsg.HasGold = true
	syncMsg.Current = "100%"				
	--send it to ourself
	self:OnChannelChat("self", syncMsg)
end

function KiruChallenge:MenuToggle()
	local visible = self.Main.Window:IsVisible()
	self.Main.Window:Show(not visible, true)
	self.DelayAutoHide = 10
	self.CombatViewState = not visible
	self.ClosedOnPurpose = visible
end

function KiruChallenge:OnCompletedEvent(challengeID)
	--remove it from the locked list if it's there
	self.DataPackage:Remove(challengeID)

	--check to make sure this challenge is in the list.
	--if it isn't, then we need to refresh the list once it is completed.
	if debugMode then self:PrintDebug("Completed Challenge ID " .. challengeID) end
	if self.Data.AllChallenges and self.Data.AllChallenges[challengeID] == nil then
		needRefresh = true
	end
	--if the challenge is on the ignore list but we just got a reward,
	--refresh the main window to show the reward button
	if self.Data.Preferences.IgnoreList ~= nil and self.Data.Preferences.IgnoreList[challengeID] ~= nil then
		needRefresh = true
	end
	
	local challenge = self.Data.AllChallenges[challengeID]
	self:PartyMsg("Challenge Finished: " .. challenge.Name)
	
	--send out the final partysync
	self:SendPartySync(challengeID,true,false)
	
	if self.Data.Preferences.AutoReward then
		--need to check the challenge in a second to make sure there's a reward.  right now it'd still be "active"
		if self.MonitorChallenge == nil then self.MonitorChallenge = {} end
		self.MonitorChallenge[challengeID] = 3
	end
end

function KiruChallenge:OnChallengeUnlocked(challenge)
	if debugMode then self:PrintDebug("Unlocked Challenge ID " .. challenge:GetId()) end
	if self.Data.AllChallenges and self.Data.AllChallenges[challengeID] == nil then
		needRefresh = true
	end
end

function KiruChallenge:OnShared()
--	Print("Shared fired")
--nothing really to do here.  the game handles the prompting already
end

function KiruChallenge:OnActivateEvent(challenge)
--	Print("Activate fired")
	if debugMode then self:PrintDebug("Something called OnActivateEvent") end
	local challengeID = challenge:GetId()
	self.StartingChallenge = nil

	self.Data.ChallengeTypes[challenge:GetType()].ActiveID = challengeID
	self.Data.ChallengeTypes[challenge:GetType()].HasActive = true
	
	self:PartyMsg("Challenge Started: " .. challenge:GetName())	
	--add to party sync
	self.PartySyncQueue[challengeID] = 1
end

function KiruChallenge:OnAbandonEvent(challengeID)
	if debugMode then self:PrintDebug("Something called OnAbandonEvent " .. challengeID) end
	--get the challenge type
	local type = self.Data.AllChallenges[challengeID].Type
	self.Data.ChallengeTypes[type].ActiveID = 0
	self.Data.ChallengeTypes[type].HasActive = false

	local challenge = self.Data.AllChallenges[challengeID]
	self:PartyMsg("Challenge Abandoned: " .. challenge.Name)	
	--send out the final partysync
	self:SendPartySync(challengeID,true,true)
end

function KiruChallenge:OnFailedEvent(challenge)
	--getting silver instead of gold is considered failing the event for some reason
	local challengeID = challenge:GetId()
	if debugMode then self:PrintDebug("Something called OnFailedEvent " .. challengeID) end
	self.Data.ChallengeTypes[challenge:GetType()].ActiveID = 0
	self.Data.ChallengeTypes[challenge:GetType()].HasActive = false
	if self.Data.AllChallenges and self.Data.AllChallenges[challengeID] == nil then
		needRefresh = true
	end
	--if the challenge is on the ignore list but we just got a reward,
	--refresh the main window to show the reward button
	if self.Data.Preferences.IgnoreList ~= nil and self.Data.Preferences.IgnoreList[challengeID] ~= nil then
		needRefresh = true
	end
	
	self:PartyMsg("Challenge Finished: " .. challenge:GetName())
	--send out the final partysync
	self:SendPartySync(challengeID,true,false)
	
	if self.Data.Preferences.AutoReward then
		--need to check the challenge in a second to make sure there's a reward.  right now it'd still be "active"
		if self.MonitorChallenge == nil then self.MonitorChallenge = {} end
		self.MonitorChallenge[challengeID] = 3
	end
end

function KiruChallenge:OnRewardListReady(challengeID,tierLevel)
	if not self.Data.Preferences.PartySpamLoot then return end
	--populate the reward list for spamming	
	local tRewardList = ChallengesLib.GetRewardList(challengeID)
	if tRewardList then 
		if not self.RewardList then self.RewardList = {} end
		for idx, tReward in ipairs(tRewardList) do
			if tReward.itemReward then
				table.insert(self.RewardList, tReward.itemReward:GetName())
			elseif tReward.monReward then
				table.insert(self.RewardList, "Renown or Coin")
			else
				table.insert(self.RewardList, "Unknown")				
			end
		end
	end
end

function KiruChallenge:OnRewardReady(challengeID,rewardIndex)
	--need to check the challenge once the reward state has cleared to remove it from the list perhaps
	if self.MonitorChallenge == nil then self.MonitorChallenge = {} end
	self.MonitorChallenge[challengeID] = 2
	
	if not self.Data.Preferences.PartySpamLoot then return end

	if self.RewardList then
		local rewardItem = self.RewardList[rewardIndex]
		local chalName = self.Data.AllChallenges[challengeID].Name
		self:PartyMsg("Got " .. rewardItem .." from " .. chalName .. ".")
		self.RewardList = nil
	end
end

function KiruChallenge:PartyMsg(message)	
	if self.Data.Preferences.PartySpam == nil or not self.Data.Preferences.PartySpam then return nil end	
	if GroupLib.InGroup() then ChatSystemLib.Command("/p " .. message) end		
end

function KiruChallenge:OnAreaRestriction(idChallenge, strHeader, strDescription, fDuration)
	if debugMode then self:PrintDebug("Something called OnAreaRestriction") end
	--if someone activated a challenge and it wasn't you, say so
	if GroupLib.InGroup() then
		if self.StartingChallenge == nil or self.StartingChallenge ~= idChallenge then
			local challenge = self.Data.AllChallenges[idChallenge]
			if challenge ~= nil then
				Print("A party member may be attempting to start " .. challenge.Name)
			end
		end
	end	
	if self.StartingChallenge ~= nil and self.StartingChallenge == idChallenge then
		ChallengesLib.ShowHintArrow(idChallenge)
	end
	if strDescription ~= nil then Print(strDescription) end
end

-- overall timer function for any updating
function KiruChallenge:OnTimerEvent()
	--if the timer fires before the player is in the world, other things could go wrong
	--skip this timer round
	if GameLib:GetPlayerUnit() == nil then return end
		
	--some initial things for the addon to load up that only needs to happen once
	if not self.InitSync then
		self.PlayerName = GameLib:GetPlayerUnit():GetName()
		self.DataPackage:SetFaction(GameLib.GetPlayerUnit():GetFaction())
		
		if self.Data.Preferences.PartySync and GroupLib.InGroup() then self:StartPartySync() end		
		self.InitSync = true
	end	
	--if we encountered a new challenge, we need to update AllChallenges to include it
	if needRefresh then
		self.Data.AllChallenges = self:PopulateAllChallenges()		
		if self.Option.Window ~= nil then self:RepopulateOptionList(true) end
		self:PopulateMainChallengeList()
		needRefresh = false
	end
	if self.StartSync ~= nil and self.StartSync > 0 then
		if debugMode then self:PrintDebug(self.StartSync) end
		if self.ChatChannel == nil then
			self.StartSync = 0
		else
			self.StartSync = self.StartSync - 1
			self:ShowSyncBar(true)
			self:UpdateSync("WAIT")
			if self.WaitSync == nil or self.WaitSync == 0 then self.WaitSync = 3 end
			if self.StartSync == 0 then
				if debugMode then self:PrintDebug("Sending Sync Request") end
				local msg = { Text = "PING" }
				local test = self.ChatChannel:SendMessage(msg)
				--if false, the channel isn't ready or nobody else is in the group
				if debugMode and test then self:PrintDebug("It thinks it sent") else self:PrintDebug("It thinks it did not send.") end		
			end		
		end
	end
	--waiting 3 seconds for a ping reply, if none, then it's not happening
	if self.WaitSync ~= nil and self.WaitSync > 0 then
		self.WaitSync = self.WaitSync - 1
		if self.WaitSync == 0 then
			self:ShowSyncBar(false)
		end	
	end
	
	if self.StopSync ~= nil and self.StopSync > 0 then
		if debugMode then self:PrintDebug(self.StopSync) end
		self.StopSync = self.StopSync - 1
		if self.StopSync == 0 then
			if debugMode then self:PrintDebug("No Reply.. stopping Sync.") end
			self.HasPartySync = false
			self:ShowSyncBar(false)
		end
	end
	if self.HideSyncAuto ~= nil and self.HideSyncAuto > 0 then
		self.HideSyncAuto = self.HideSyncAuto - 1
		if self.PartySync.HideButton ~= nil then 
			if self.HideSyncAuto < (self.Data.Preferences.SyncAutoHideTime - 4) then
				self.PartySync.HideButton:SetText("Hide in " .. self.HideSyncAuto)
			else
				self.PartySync.HideButton:SetText("Hide")			
			end 
		end
		if self.HideSyncAuto == 0 then
			self:HideSync()
		end	
	end
	
	--if someone takes a reward for a challenge that shouldnt be on the list, remove it
	if self.MonitorChallenge ~= nil and self:GetTableLength(self.MonitorChallenge) > 0 then
		for key, val in pairs(self.MonitorChallenge) do
			local shouldShow = self:ShouldDisplayChallenge(key)
			if val == 2 then --check for reward state to remove
				if shouldShow ~= 2 then
					--no longer under reward state, remove it from the list if needed
					if shouldShow ~= 1 then self:RemoveFromMainChallengeList(key) end
					--also if options are open, refresh the @ completed
					if self.Option.Window ~= nil then self:RepopulateOptionList(true) end
					--drop it from monitoring
					self.MonitorChallenge[key] = nil
				end 
			end
			if val == 3 or val == 4 then --auto loot check for one second
				if shouldShow == 2 then
					Event_FireGenericEvent("ChallengeRewardShow", key)
					val = 5
				else
					self.MonitorChallenge[key] = val + 1
				end
			end
			if val == 5 then --last check auto reward must not have had a reward.  remove it from monitoring
				self.MonitorChallenge[key] = nil			
			end
		end
	end
			
	local results = self:ProcessChallengeList()
	--if we have PartySync on and there are challenges in process, send out the syncs
	if self.Data.Preferences.PartySync and (self.HasPartySync or partyOfOne) then
		if self:GetTableLength(self.PartySyncQueue) > 0 then
			for key, val in pairs(self.PartySyncQueue) do
				self:SendPartySync(key,false,false)
			end
		end
	else
		--if there's anything in the queue, delete it
		if self:GetTableLength(self.PartySyncQueue) > 0 then self.PartySyncQueue = {} end
	end
	
	--if the window was closed on purpose, no point deciding whether to hide it or show it
	if self.ClosedOnPurpose then return end
	
	if results.Total == 0 then
		--hide the window if that option is on
		if self.Data.Preferences.HideNoChallenges then 
			--unless we have a delay
			if self.DelayAutoHide ~= nil and self.DelayAutoHide > 0 then
				self.DelayAutoHide = self.DelayAutoHide - 1
			else
				self.Main.Window:Show(false,true)
				self.CombatViewState = false		
			end
		else
			if self.Data.Preferences.HideCombat == true and self.InCombat == true then
				self.CombatViewState = true
			else
				self.Main.Window:Show(true,true)
			end
			self.CombatViewState = true
			self:SetFormBGSprite("Gray")
		end
	else
		--make sure the window is shown
		if self.Data.Preferences.HideNoChallenges then 
			if self.Data.Preferences.HideCombat == true and self.InCombat == true then
				self.CombatViewState = true
			else
				self.Main.Window:Show(true,true)
			end
			self.CombatViewState = true
		end
		
		--pick a color for the border
		if results.Active > 0 then
			self:SetFormBGSprite("White")
		elseif results.Okay > 0 then
			self:SetFormBGSprite("Green")
		else
			self:SetFormBGSprite("Blue")				
		end
	end
end

function KiruChallenge:SendPartySync(challengeID,isFinal,abandoned)
	if (self.Data.Preferences.PartySync and self.HasPartySync and self.ChatChannel ~= nil)
	or (self.Data.Preferences.PartySync and partyOfOne) then
	
		local challenge = self.Data.AllChallenges[challengeID]
		if challenge == nil then return nil end
		
		if debugMode then self:PrintDebug("Preparing Sync message") end
		local syncMsg = {}
		syncMsg.Text = "UPDATE"
		if isFinal then syncMsg.Text = "FINAL" end
		syncMsg.Player = self.PlayerName
		syncMsg.CID = challengeID
		syncMsg.CName = challenge.Name
		local sendStatus = 1
		if abandoned then sendStatus = 2 end
		syncMsg.Status = sendStatus
		syncMsg.Tier = challenge.SourceChallenge:GetCurrentTier()
		syncMsg.DTier = challenge.SourceChallenge:GetDisplayTier()
		syncMsg.Time = self:HelperConvertToTime(challenge.SourceChallenge:GetTimer())
		if challenge.TierCounts == nil then
			challenge.IsTimeTiered = challenge.SourceChallenge:IsTimeTiered()
			challenge.TierCounts = challenge.SourceChallenge:GetAllTierCounts()
		end
		syncMsg.IsTime = challenge.IsTimeTiered
		--some challenges don't have medals, just a single tier of completion
		syncMsg.HasGold = true
		if challenge.TierCounts[3] == nil then syncMsg.HasGold = false end
		--calc percentages			
		local tierCounts = challenge.TierCounts
		local currentCount = challenge.SourceChallenge:GetCurrentCount()
		if not syncMsg.IsTime and syncMsg.Tier < 3 then
			local tier = challenge.TierCounts[syncMsg.Tier + 1].nGoalCount
			syncMsg.Current = tostring(self:Round((currentCount * 100) / tier, 0)) .. "%"				
		else
			syncMsg.Current = "0d"
		end
		if debugMode then self:PrintDebug(self:PrintVariable(syncMsg,"SyncMsg",0,0,3)) end
		
		--send the message out
		if self.HasPartySync and self.ChatChannel ~= nil then
			self.ChatChannel:SendMessage(syncMsg)
		end

		--send it to ourself too so we're on the list
		self:OnChannelChat("self", syncMsg)
	end
	--if this is the final message (completed the event somehow) then remove it from the queue
	if isFinal and self.PartySyncQueue[challengeID] ~= nil then self.PartySyncQueue[challengeID] = nil end		
end

-----------------------------------------------------------------------------------------
--Data Loading and Parsing
-----------------------------------------------------------------------------------------
-- load all challenge data into a table
function KiruChallenge:PopulateAllChallenges()
	local tempTable = {}
	
	if debugMode then self:PrintDebug("Generating All Challenges Table") end
	
	if GameLib:GetPlayerUnit() and ChallengesLib.GetActiveChallengeList() ~= nil then	
		for key, val in pairs(ChallengesLib.GetActiveChallengeList()) do			
			local challengeObject = self:CreateChallengeObject(val)
			
			if challengeObject == nil then break end
			
			--store the object using its ID as a key so we can retrieve it easier later
			tempTable[challengeObject.ID] = challengeObject 
			
			--remove it from the locked list
			self.DataPackage:Remove(challengeObject.ID)
			
			--if the challenge is active, store its type
			if challengeObject.SourceChallenge:IsActivated() then
				self.Data.ChallengeTypes[challengeObject.Type].ActiveID = challengeObject.ID
				self.Data.ChallengeTypes[challengeObject.Type].HasActive = true
			end	
		end
				
		if debugMode then self:PrintDebug("Loaded Challenge Total: " .. self:GetTableLength(tempTable))	end		
	end	
	
	return tempTable
end

function KiruChallenge:DoubleCheckChallengeList()
	--if AllChallenges is empty, try refilling it once.  if it fails again there must be a problem
	--or the player simply has not unlocked any challenges yet
	if self:GetTableLength(self.Data.AllChallenges) == 0 then
		self.Data.AllChallenges = self:PopulateAllChallenges()
		if self:GetTableLength(self.Data.AllChallenges) == 0 then return 0 end		
	end
	return 1
end

function KiruChallenge:PopulateOptionList(reloadCompleted)
	if self:DoubleCheckChallengeList() == 0 then return end	
	--populates the Option form with all challenges and flags them for ignoring or whitelisting.
	--make sure the option lists are empty to avoid doubles
	self:ClearOptionList()
	
	--sort AllChallenges by Zone, then by Challenge Name
	local keySort = self:GetKeysSortedByZoneThenName(self.Data.AllChallenges)

	--set up for sorting and the zone headers	
	local sortOrder = 0
	local lastZone = ""
	
	--useless metric
	local completed = 0
	
	--build the children objects, embedding their sort order into their data
	for _, key in ipairs(keySort) do
		local challenge = self.Data.AllChallenges[key]
		--first see if a zone header needs to be built
		if challenge.ZoneName ~= lastZone then
			sortOrder = sortOrder + 1
			lastZone = challenge.ZoneName
			local showText = "Zone: " .. lastZone
			self.Option.ItemList.Items[key .. "h"] = self:CreateOptionHeader(showText, showText, sortOrder, false, false)
			self.Option.ItemListWhite.Items[key .. "h"] = self:CreateOptionHeader(showText, showText, sortOrder, true, false)
			self.Option.AutoStartList.Items[key .. "h"] = self:CreateOptionHeader(showText, showText, sortOrder, false, false, true)
		end
		--add the listitem for the challenge
		sortOrder = sortOrder + 1
		self.Option.ItemList.Items[key] = self:CreateOptionListItem(key, sortOrder, false)
		self.Option.ItemListWhite.Items[key] = self:CreateOptionListItem(key, sortOrder, true)
		self.Option.AutoStartList.Items[key] = self:CreateOptionListItem(key, sortOrder, false, false, true)
		--sum up the # completed challenges
		if reloadCompleted then
			challenge.TotalCompleted = challenge.SourceChallenge:GetCompletionCount()
		end
		local numComplete = challenge.TotalCompleted
		if numComplete > 0 then completed = completed + numComplete end
	end	
	
	-- now all the item are added and sorted, call ArrangeChildrenVert to list out the list items vertically
	--ArrangeChildren(location,sortOrder)
	--location.. 0=top, 1=centered, 2=bottom
	--sortOrder.. standard sorting function
	--returns nChildrenHeight - vertical size occupied by the children
	self.Option.ItemList.Window:ArrangeChildrenVert(0, function(a,b)
			local idA = a:GetData()
			local idB = b:GetData()			
			return idA.SortOrder < idB.SortOrder
		end
	)
	--repeat for the white list
	self.Option.ItemListWhite.Window:ArrangeChildrenVert(0, function(a,b)
			local idA = a:GetData()
			local idB = b:GetData()			
			return idA.SortOrder < idB.SortOrder
		end
	)	
	
	--repeat for the autostart list
	self.Option.AutoStartList.Window:ArrangeChildrenVert(0, function(a,b)
			local idA = a:GetData()
			local idB = b:GetData()			
			return idA.SortOrder < idB.SortOrder
		end
	)	
	return completed
end

function KiruChallenge:PopulateOptionLockedList()
	self:ClearOptionLockedList()

	if self.DataPackage:GetCount() == 0 then return end	
	
	--sort the locked by Zone, then by Challenge Name
	local locked = self.DataPackage:GetList()
	local keySort = self:GetKeysSortedByZoneThenName(locked)

	--set up for sorting and the zone headers	
	local sortOrder = 0
	local lastZone = ""
	--build the children objects, embedding their sort order into their data
	for _, key in ipairs(keySort) do
		local challenge = locked[key]
		--first see if a zone header needs to be built
		if challenge.ZoneName ~= lastZone then
			sortOrder = sortOrder + 1
			lastZone = challenge.ZoneName
			local showText = "Zone: " .. lastZone
			self.Option.LockedList.Items[key .. "h"] = self:CreateOptionHeader(showText, showText, sortOrder, false, true)
		end
		--add the listitem for the challenge
		sortOrder = sortOrder + 1
		self.Option.LockedList.Items[key] = self:CreateOptionListItem(key, sortOrder, false, true)
	end	
	
	self.Option.LockedList.Window:ArrangeChildrenVert(0, function(a,b)
			local idA = a:GetData()
			local idB = b:GetData()			
			return idA.SortOrder < idB.SortOrder
		end
	)
end

function KiruChallenge:ShouldDisplayChallenge(challenge)
	--should display a challenge if one of these is true:
	--  1. the challenge is in the zone and not ignored
	--  2. the challenge has a reward
	--  3. the challenge is on the whitelist
	
	--if an ID is passed instead of a full challenge, get it
	if string.lower(type(challenge)) ~= "table" then challenge = self.Data.AllChallenges[challenge] end
	if challenge == nil then return 0 end
	
	--check if it has a reward pending.. if so, we're done checking
	if challenge.SourceChallenge:ShouldCollectReward() then return 2 end

	--check if its on the whitelist.. if it is, we're done checking
	if self.Data.Preferences.WhiteList[challenge.ID] ~= nil then return 1 end	
	
	--otherwise check if it's in the same zone and not being ignored
	if GameLib.IsInWorldZone(challenge.ZoneID) then
	
		--are we ignoring it?
		if self.Data.Preferences.IgnoreList[challenge.ID] ~= nil then return 3 end
		
		--good to show
		return 1
	end

	--nope, no reason to show it
	return 0	
end

function KiruChallenge:PopulateMainChallengeList()
	--scans through self.Data.AllChallenges picking out the local ones for the zone the player is in.
	--happens when the player changes SubZone areas (more often than you might think actually)
	--adds or removes list items as needed, then sorts		
	if self:DoubleCheckChallengeList() == 0 then return end
	
	--populate the Local table with local challenges, or ones that are ready for reward
	local a = 0
	local d = 0
	local e = 0
	local i = 0
	
	for _, val in pairs(self.Data.AllChallenges) do
	
		local shouldShow = self:ShouldDisplayChallenge(val)
	
		if shouldShow == 1 or shouldShow == 2 then
			if self.Main.ItemList.Items == nil or self.Main.ItemList.Items[val.ID] == nil then
				self.Main.ItemList.Items[val.ID] = self:CreateChallengeListItem(val.ID)
				a = a + 1
			else
				e = e + 1
			end			
		elseif shouldShow == 3 then
			i = i + 1
		else
			--not to be shown.  if it's in the list, remove it
			if self.Main.ItemList.Items and self.Main.ItemList.Items[val.ID] ~= nil then 
				d = d + 1
				--delete the window item first
				self.Main.ItemList.Items[val.ID].Window:Destroy()
				self.Main.ItemList.Items[val.ID] = nil				
			end
		end
	end  --end for
		
	if debugMode then self:PrintDebug("Main List Updated: Total " .. self:GetTableLength(self.Main.ItemList.Items) .. " A: " .. a .. " E: " .. e .. " D: " .. d .. " I: " .. i ) end
end

function KiruChallenge:GetChallengeStatus(challengeID)
	if challengeID == nil then return nil end

	local challenge = self.Data.AllChallenges[challengeID]

	if challenge == nil then return nil end	
	
	local returnStatus = {}

	--if there is a reward pending...
	if challenge.SourceChallenge:ShouldCollectReward() then
		returnStatus.Status = Status_Reward		
		return returnStatus	
	end
	
	--if this challenge is active already..
	if challenge.SourceChallenge:IsActivated() then
		returnStatus.Status = Status_Active		
		return returnStatus	
	end
	
	--if this challenge is on Cooldown..
	if challenge.SourceChallenge:IsInCooldown() then
		returnStatus.Status = Status_Cooling
		returnStatus.Cooldown = self:HelperConvertToTime(challenge.SourceChallenge:GetTimer())
		returnStatus.Distance = self:GetChallengeDistance(challengeID)	
		return returnStatus	
	end
	
	--begin checking for restrictions and player locations and whatnot
	--if its a housing challenge, do we know if it's in the current zone?
	if challenge.IsHouse then
		if not self.IsInHousing or self.CurrentHouse == nil or self.CurrentHouse[challengeID] == nil then 
			returnStatus.Status = Status_Housing 
			return returnStatus
		end
	end
		
	--if there is a zone restriction, easy stopping point
	if challenge.ZoneRestriction then
		local zoneID = challenge.ZoneRestriction.idSubZone
		local zoneName = challenge.ZoneRestriction.strSubZoneName
	
		--if zoneID is 0, there's no restriction.  otherwise, check the player's location
		if not (zoneID == 0) and not (GameLib.IsInWorldZone(zoneID)) then
			returnStatus.Status = Status_Area
			returnStatus.Distance = self:GetChallengeDistance(challengeID)
			return returnStatus				
		end
	end
	
	--if there is a start location ID and the player isn't there.. easy stopping point
	if challenge.StartLocationID then	
		--if zoneID is 0, there's no restriction.  otherwise, check the player's location
		if not (challenge.StartLocationID == 0) and not (GameLib.IsInLocation(challenge.StartLocationID)) then
			returnStatus.Status = Status_Area
			returnStatus.Distance = self:GetChallengeDistance(challengeID)
			return returnStatus				
		end
	end
	
	--not sure what this is, but its in Carbine's.. I guess this is for challenges
	--that can only be done a certain number of times?
	if challenge.TotalCompleted then	
		local maxTotal = challenge.SourceChallenge:GetCompletionTotal()
		--if CompletionTotal is -1, there's no restriction.  otherwise, check against total completed
		if not (maxTotal == -1) and not (challenge.TotalCompleted < maxTotal) then
			returnStatus.Status = Status_NoMore		
			return returnStatus				
		end
	end
	
	--are we already running a challenge of this type?
	if self.Data.ChallengeTypes[challenge.Type] ~= nil and self.Data.ChallengeTypes[challenge.Type].HasActive then
		local activeID = self.Data.ChallengeTypes[challenge.Type].ActiveID	
		if activeID ~= challengeID then
			--is it still running?
			if self:GetChallengeStatus(activeID).Status == Status_Active then
				returnStatus.Status = Status_Wait
				if challenge.IsHouse and self.CurrentHouse[challengeID] ~= nil then
					returnStatus.Distance = self:GetChallengeDistance(challengeID,self.CurrentHouse[challengeID].Pos)
				else
					returnStatus.Distance = self:GetChallengeDistance(challengeID) 
				end
				return returnStatus
			else
				--false positive, clean it up
				self.Data.ChallengeTypes[challenge.Type].ActiveID = 0
				self.Data.ChallengeTypes[challenge.Type].HasActive = false
			end			
		end
	end
	
	--if we get to this point, it "should" be okay to start this Challenge
	returnStatus.Status = Status_OK
	if challenge.IsHouse and self.CurrentHouse[challengeID] ~= nil then
		returnStatus.Distance = self:GetChallengeDistance(challengeID,self.CurrentHouse[challengeID].Pos)
	else
		returnStatus.Distance = self:GetChallengeDistance(challengeID) 
	end
	return returnStatus				
end

-----------------------------------------------------------------------------------------------
-- ItemList Functions
-----------------------------------------------------------------------------------------------

function KiruChallenge:ShowChallengeItem(item,show)
	if item.Visibility and item.Visibility == show then return end
	item.Window:Show(show,true)
	item.Visibility = show
end

-- populate item list
function KiruChallenge:ProcessChallengeList()	
	
	--return a table of totals for further window processing
	local returnListTotals = {}
	returnListTotals.Total = self:GetTableLength(self.Main.ItemList.Items)
	
	self.KeybindChallenge = nil
	
	if returnListTotals.Total > 0 then
		-- sorting the challenges into containers based on status
		local rewardTable = {}
		local okayTable = {}
		local activeTable = {}
		local areaTable = {}
		local coolingTable = {}
		local otherTable = {}
		
		--loop through the challenges, update their colours and text
		--based on current status, then add them into their container
		for idx, chalListItem in pairs(self.Main.ItemList.Items) do
			local challenge = self.Data.AllChallenges[chalListItem.ChallengeID]			
			local checkStatus = self:GetChallengeStatus(chalListItem.ChallengeID)
			local status = checkStatus.Status

			chalListItem.Status:SetText(status)
			
			if status == Status_Reward then
				chalListItem.Title:SetTextColor(Status_Reward_Color)
				chalListItem.Status:SetTextColor(Status_Reward_Color)
				--insert into reward table with challenge name for sorting
				rewardTable[chalListItem.ChallengeID] = challenge.Name
				self:ShowChallengeItem(chalListItem,true)
			elseif status == Status_Active then
				chalListItem.Title:SetTextColor(Status_Active_Color)
				chalListItem.Status:SetTextColor(Status_Active_Color)
				--insert into okay table with challenge name for sorting
				activeTable[chalListItem.ChallengeID] = challenge.Name
				self:ShowChallengeItem(chalListItem,true)
			elseif status == Status_OK then
				chalListItem.Title:SetTextColor(Status_OK_Color)
				chalListItem.Status:SetTextColor(Status_OK_Color)
				if checkStatus.Distance == nil then checkStatus.Distance = 30000 end
				if checkStatus.Distance == -1 then checkStatus.Distance = 30000	end
				if (checkStatus.Distance ~= 30000) and (self.Data.Preferences.WhiteList[chalListItem.ChallengeID] == nil) and self.Data.Preferences.FilterDistance < 2000 and checkStatus.Distance > self.Data.Preferences.FilterDistance then
					self:ShowChallengeItem(chalListItem,false)			
				else
					self:ShowChallengeItem(chalListItem,true)			
					--insert into okay table with challenge name for sorting
					if self.Data.Preferences.ReadyDistance then
						if checkStatus.Distance ~= 30000 then chalListItem.Status:SetText(checkStatus.Distance .. "m") end
						okayTable[chalListItem.ChallengeID] = { A = checkStatus.Distance, B = challenge.Name, C = 1, D = checkStatus.Distance }							
					else
						okayTable[chalListItem.ChallengeID] = { A = challenge.Name , B = checkStatus.Distance, C = 1, D = checkStatus.Distance }			
					end
				end
			elseif status == Status_Wait then
				--type is Okay but there's another of this type active, so waiting
				chalListItem.Title:SetTextColor(Status_Wait_Color)
				chalListItem.Status:SetTextColor(Status_Wait_Color)
				if checkStatus.Distance == nil then checkStatus.Distance = 30000 end
				if checkStatus.Distance == -1 then checkStatus.Distance = 30000	end
				if (checkStatus.Distance ~= 30000) and (self.Data.Preferences.WhiteList[chalListItem.ChallengeID] == nil) and self.Data.Preferences.FilterDistance < 2000 and checkStatus.Distance > self.Data.Preferences.FilterDistance then
					self:ShowChallengeItem(chalListItem,false)			
				else
					self:ShowChallengeItem(chalListItem,true)			
					--insert into okay table with challenge name for sorting
					if self.Data.Preferences.ReadyDistance then
						if checkStatus.Distance ~= 30000 then chalListItem.Status:SetText(checkStatus.Distance .. "m") end
						okayTable[chalListItem.ChallengeID] = { A = checkStatus.Distance, B = challenge.Name, D = checkStatus.Distance }							
					else
						okayTable[chalListItem.ChallengeID] = { A = challenge.Name , B = checkStatus.Distance, D = checkStatus.Distance }			
					end
				end
			elseif status == Status_Area then
				chalListItem.Title:SetTextColor(Status_Area_Color)
				chalListItem.Status:SetTextColor(Status_Area_Color)
				if checkStatus.Distance == nil then checkStatus.Distance = 30000 end
				if checkStatus.Distance == -1 then
					chalListItem.Status:SetText("???m")
					checkStatus.Distance = 30000			
				else
					chalListItem.Status:SetText(checkStatus.Distance .. "m")			
				end
				if (checkStatus.Distance ~= 30000) and (self.Data.Preferences.WhiteList[chalListItem.ChallengeID] == nil) and self.Data.Preferences.FilterDistance < 2000 and checkStatus.Distance > self.Data.Preferences.FilterDistance then
					self:ShowChallengeItem(chalListItem,false)			
				else
					self:ShowChallengeItem(chalListItem,true)			
					--insert into area table with challenge distance for sorting
					areaTable[chalListItem.ChallengeID] = { A = checkStatus.Distance, B = challenge.Name }
				end
				--insert into area table with challenge distance for sorting
			elseif status == Status_Housing then
				chalListItem.Title:SetTextColor(Status_Housing_Color)
				chalListItem.Status:SetTextColor(Status_Housing_Color)
				chalListItem.Status:SetText("Housing")
				if (self.Data.Preferences.WhiteList[chalListItem.ChallengeID] == nil) and self.Data.Preferences.FilterDistance < 2000 then
					self:ShowChallengeItem(chalListItem,false)			
				else
					self:ShowChallengeItem(chalListItem,true)			
					--insert into area table with challenge distance for sorting
					areaTable[chalListItem.ChallengeID] = { A = 30000, B = challenge.Name }
				end
			elseif status == Status_Cooling then
				if (self.Data.Preferences.ShowCooldown == true) then
					chalListItem.Title:SetTextColor(Status_Cooling_Color)
					chalListItem.Status:SetTextColor(Status_Cooling_Color)
					chalListItem.Status:SetText(checkStatus.Cooldown)
					if checkStatus.Distance == nil then checkStatus.Distance = 30000 end
					if checkStatus.Distance == -1 then checkStatus.Distance = 30000	end
					if (checkStatus.Distance ~= 30000) and (self.Data.Preferences.WhiteList[chalListItem.ChallengeID] == nil) and self.Data.Preferences.FilterDistance < 2000 and checkStatus.Distance > self.Data.Preferences.FilterDistance then
						self:ShowChallengeItem(chalListItem,false)			
					else
						self:ShowChallengeItem(chalListItem,true)			
						--insert into cooling table with the cooldown time
						coolingTable[chalListItem.ChallengeID] = checkStatus.Cooldown			
					end
				else
					self:ShowChallengeItem(chalListItem,false)
				end
			else
				--insert into other table with challenge name for sorting
				otherTable[chalListItem.ChallengeID] = challenge.Name			
			end
		end
		
		local sortOrder = 0
		
		--apply the sort order to the window items in order of importance
		returnListTotals.Active = self:GetTableLength(activeTable)
		if returnListTotals.Active > 0 then
			--sort by value (challenge name in this case) and add to list
			local keySort = self:GetKeysSortedByValue(activeTable)
			for key, val in ipairs(keySort) do
				sortOrder = sortOrder + 1
				self:ShowReadyArrow(self.Main.ItemList.Items[val], false)
				self.Main.ItemList.Items[val].SortOrder = sortOrder				
			end				
		end
		
		returnListTotals.Reward = self:GetTableLength(rewardTable)
		returnListTotals.Okay = self:GetTableLength(okayTable)
		returnListTotals.Area = self:GetTableLength(areaTable)
		returnListTotals.Cooling = self:GetTableLength(coolingTable)
		
		returnListTotals.Total = returnListTotals.Active + returnListTotals.Reward + returnListTotals.Okay + returnListTotals.Area + returnListTotals.Cooling
		
		local keybindOrder = sortOrder + self.KeybindShiftDown
		if (keybindOrder + 1) > (returnListTotals.Total - returnListTotals.Active - returnListTotals.Cooling) then 
			self.KeybindShiftDown = 0
			keybindOrder = sortOrder 
		end
		if self.KeybindShiftDown < 0 then
			self.KeybindShiftDown = (returnListTotals.Total - returnListTotals.Active - returnListTotals.Cooling) - 1
			keybindOrder = (returnListTotals.Total - returnListTotals.Active - returnListTotals.Cooling) - 1
		end	
	
		if returnListTotals.Reward > 0 then
			--sort by value (challenge name in this case) and add to list
			local keySort = self:GetKeysSortedByValue(rewardTable)
			for key, val in ipairs(keySort) do
				sortOrder = sortOrder + 1
				self:ShowReadyArrow(self.Main.ItemList.Items[val], false)
				self.Main.ItemList.Items[val].SortOrder = sortOrder	
				if sortOrder == keybindOrder + 1 then 
					self.KeybindChallenge = val 
					self:ShowKeybindNub(self.Main.ItemList.Items[val], true)
					self:HideOtherNubs(val)
				end			
			end				
		end
		if returnListTotals.Okay > 0 then
			--sort by value (challenge name in this case) and add to list
			keySort = self:GetKeysSortedByAThenB(okayTable)
			for key, val in ipairs(keySort) do
				sortOrder = sortOrder + 1
				self:ShowReadyArrow(self.Main.ItemList.Items[val], self.Data.Preferences.ReadyDistance)
				self.Main.ItemList.Items[val].SortOrder = sortOrder	
				if self.Data.Preferences.AutoStart[val] ~= nil and self.Data.Preferences.AutoStart[val].O == 1 and okayTable[val].C == 1 then
					local distance = self.Data.Preferences.AutoStart[val].D or 10
					if okayTable[val].D ~= nil and (okayTable[val].D < (distance + 1)) then self:TriggerStartChallenge(val) end
				end		
				if sortOrder == keybindOrder + 1 then 
					self.KeybindChallenge = val 
					self:ShowKeybindNub(self.Main.ItemList.Items[val], true)
					self:HideOtherNubs(val)
				end			
			end				
		end
		if returnListTotals.Area > 0 then
			--sort by value (distance in this case)
			local keySort = self:GetKeysSortedByAThenB(areaTable)
			for key, val in ipairs(keySort) do
				sortOrder = sortOrder + 1
				self:ShowReadyArrow(self.Main.ItemList.Items[val], false)
				self.Main.ItemList.Items[val].SortOrder = sortOrder				
				if sortOrder == keybindOrder + 1 then 
					self.KeybindChallenge = val 
					self:ShowKeybindNub(self.Main.ItemList.Items[val], true)
					self:HideOtherNubs(val)
				end			
			end				
		end
		if returnListTotals.Cooling > 0 then
			--sort by value (challenge name in this case) and add to list
			local keySort = self:GetKeysSortedByValue(coolingTable)
			for key, val in ipairs(keySort) do
				sortOrder = sortOrder + 1
				self:ShowReadyArrow(self.Main.ItemList.Items[val], false)
				self.Main.ItemList.Items[val].SortOrder = sortOrder				
				if sortOrder == keybindOrder + 1 then 
					self.KeybindChallenge = val 
					self:ShowKeybindNub(self.Main.ItemList.Items[val], true)
					self:HideOtherNubs(val)
				end			
			end				
		end

		-- now all the item are added, call ArrangeChildrenVert to list out the list items vertically
		--ArrangeChildren(location,sortOrder)
		--location.. 0=top, 1=centered, 2=bottom
		--sortOrder.. standard sorting function
		--returns nChildrenHeight - vertical size occupied by the children
		self.Main.ItemList.Window:ArrangeChildrenVert(0, function(a,b)
			local idA = a:GetData()
			local idB = b:GetData()
			
			return self.Main.ItemList.Items[idA].SortOrder < self.Main.ItemList.Items[idB].SortOrder
			end
		)
	end	
		
	--return the list of totals for further window processing
	return returnListTotals	
end

-- clear the item lists
function KiruChallenge:ClearChallengeList()
	if self.Main.ItemList.Window ~= nil then self.Main.ItemList.Window:DestroyChildren() end
	-- clear the list item array
	self.Main.ItemList.Items = {}
end

-- clear the option list
function KiruChallenge:ClearOptionList()
	if self.Option == nil or self.Option.ItemList == nil then return end
	
	-- destroy all the wnd inside the lists
	if self.Option.ItemList.Window ~= nil then self.Option.ItemList.Window:DestroyChildren() end
	if self.Option.ItemListWhite.Window ~= nil then self.Option.ItemListWhite.Window:DestroyChildren() end
	if self.Option.AutoStartList.Window ~= nil then self.Option.AutoStartList.Window:DestroyChildren() end
	--clear the list array
	self.Option.ItemList.Items = {}
	self.Option.ItemListWhite.Items = {}
	self.Option.AutoStartList.Items = {}
end

-- clear the option locked list
function KiruChallenge:ClearOptionLockedList()
	if self.Option == nil or self.Option.LockedList == nil then return end
	
	-- destroy all the wnd inside the lists
	if self.Option.LockedList.Window ~= nil then self.Option.LockedList.Window:DestroyChildren() end
	--clear the list array
	self.Option.LockedList.Items = {}
end

-- clear the partysync list
function KiruChallenge:ClearPartySyncList()
	if self.PartySync == nil or self.PartySync.ItemList == nil then return end
	if self.PartySync.ItemList.Window ~= nil then self.PartySync.ItemList.Window:DestroyChildren() end
	self.PartySync.ItemList.Items = {}
end

-- create a challenge list item in the parent then returns the window objects
function KiruChallenge:CreateChallengeListItem(challengeID)
	if challengeID == nil then return nil end
		
	local challenge = self.Data.AllChallenges[challengeID]
	
	if challenge == nil then return nil end
		
	-- load the window item for the list item
	local wnd = Apollo.LoadForm(self.xmlDoc, "ChallengeListItem", self.Main.ItemList.Window, self)	

	wnd:SetData(challenge.ID)
	wnd:SetTooltip(challenge.Name .. ":\n" .. challenge.Description)

	local wndTitle = wnd:FindChild("Title")
	local wndStatus = wnd:FindChild("Status")
	local wndArrow = wnd:FindChild("Arrow")
	wndArrow:SetData(challenge.ID)
	
	--keybind nub
	wndKeybind = wnd:FindChild("KeybindNub")
	wndKeybind:Show(false,true)
		
	if wndTitle then wndTitle:SetText(challenge.Name) end
	if wndStatus then wndStatus:SetText("Init")	end	

	--wrap it up
	local storedWindow = {}
	storedWindow.Window = wnd
	storedWindow.Title = wndTitle
	storedWindow.Status = wndStatus
	storedWindow.ChallengeID = challenge.ID
	storedWindow.KeybindNub = wndKeybind
	storedWindow.Arrow = wndArrow
	storedWindow.NubShowing = false
	storedWindow.SortOrder = 0 --will be updated later
			
	if debugMode then self:PrintDebug("Added ID: " .. challenge.ID .. " " .. challenge.Name) end
	
	return storedWindow
end

function KiruChallenge:ShowKeybindNub(item,show)
	if not self.Data.Preferences.ShowKeybindNub then show = false end
	if self.Keybind == nil then show = false end
	local l,t,r,b = item.Title:GetAnchorOffsets()
	if show then
		--move the title over to 31
		if l ~= 31 then item.Title:SetAnchorOffsets(31,t,r,b) end
		local keytext = tblSpecialKeys[self.Keybind] or string.char(self.Keybind)
		item.KeybindNub:SetText(keytext)
		
		--scroll the list if needed
		if self.KeybindShiftOnce then
			l,t,r,b = item.Window:GetAnchorOffsets()
			local bottomScroll = self.Data.Preferences.ShowScroll * 20
			local scrollPos = self.Main.ItemList.Window:GetVScrollPos()
			if t < scrollPos then self.Main.ItemList.Window:SetVScrollPos(t) end
			if t >= bottomScroll then self.Main.ItemList.Window:SetVScrollPos(t) end
			self.KeybindShiftOnce = false
		end
	else
		--move the title back to 5
		if l ~= 5 then item.Title:SetAnchorOffsets(5,t,r,b) end
	end
	item.KeybindNub:Show(show, true)
	item.NubShowing = show
end

function KiruChallenge:HideOtherNubs(idx)
	for key, val in pairs(self.Main.ItemList.Items) do
		if key ~= idx and val.NubShowing then
			self:ShowKeybindNub(val,false)
		end
	end
end

function KiruChallenge:ShowReadyArrow(item,show)
	if not self.Data.Preferences.ReadyDistance then show = false end
	local l,t,r,b = item.Status:GetAnchorOffsets()
	if show then
		if r ~= -15 then item.Status:SetAnchorOffsets(-68,t,-18,b) end		
	else
		if r ~= 0 then item.Status:SetAnchorOffsets(-50,t,0,b) end
	end
	item.Arrow:Show(show, true)
end

-- create an option list item in the parent then returns the window objects
function KiruChallenge:CreateOptionListItem(challengeID, sortOrder, isWhiteList, isLockedList, isAutoStartList)
	if challengeID == nil then return nil end
		
	local challenge = self.Data.AllChallenges[challengeID]
	if isLockedList then challenge = self.DataPackage:GetChallenge(challengeID) end
	
	if challenge == nil then return nil end
	
	local wnd = nil
	-- load the window item for the list item
	if isWhiteList then	wnd = Apollo.LoadForm(self.xmlDoc, "OptionListItem", self.Option.ItemListWhite.Window, self) end
	if isLockedList then wnd = Apollo.LoadForm(self.xmlDoc, "OptionListItem", self.Option.LockedList.Window, self) end
	if isAutoStartList then wnd = Apollo.LoadForm(self.xmlDoc, "OptionListItem", self.Option.AutoStartList.Window, self) end
	if (not isWhiteList) and (not isLockedList) and (not isAutoStartList) then wnd = Apollo.LoadForm(self.xmlDoc, "OptionListItem", self.Option.ItemList.Window, self) end
	
	local data = {ChallengeID = challengeID, SortOrder = sortOrder, IsWhiteList = isWhiteList, IsAutoStart = isAutoStartList}
	wnd:SetData(data)
	if not isLockedList then wnd:SetTooltip(challenge.Name .. ":\n" .. challenge.Description) end

	local wndBtnToggle = wnd:FindChild("Button_Toggle")
	local wndBtnEdit = wnd:FindChild("Button_Edit")
	
	if isLockedList then
		wndBtnToggle:Show(false,true)
		wndBtnEdit:Show(false, true)
	else
		--check if it was previously set
		if isWhiteList then
			if self.Data.Preferences.WhiteList[challengeID] ~= nil then wndBtnToggle:SetCheck(true) end	
			wndBtnEdit:Show(false, true)
		else
			if isAutoStartList then
				if self.Data.Preferences.AutoStart[challengeID] ~= nil and self.Data.Preferences.AutoStart[challengeID].O == 1 then wndBtnToggle:SetCheck(true) end
				wndBtnEdit:Show(true, true)
				wndBtnEdit:SetData(challengeID)	
			else
				if self.Data.Preferences.IgnoreList[challengeID] ~= nil then wndBtnToggle:SetCheck(true) end		
				wndBtnEdit:Show(false, true)
			end
		end	
	end
	
	local wndTitle = wnd:FindChild("Title")
	local wndStatus = wnd:FindChild("Status")
	
	if wndTitle then
		--scoot it over if no button
		if isLockedList then
			local l, t, r, b = wndTitle:GetAnchorOffsets()
			l = 5
			wndTitle:SetAnchorOffsets(l,t,r,b)
		end
		wndTitle:SetText(challenge.Name)
		wndTitle:SetTextColor("FFFFFFFF")
	end
	if wndStatus then	
		if isLockedList then
			local l, t, r, b = wndStatus:GetAnchorOffsets()
			l = l - 14
			wndStatus:SetAnchorOffsets(l,t,r,b)
			if challenge.x ~= 0 and challenge.z ~= 0 then wndStatus:SetText(challenge.x .. "," .. challenge.z) end
		else
			if isAutoStartList then
				wndStatus:Show(false)
			else
				if challenge.MaxCompleted ~= -1 then
					wndStatus:SetText(challenge.TotalCompleted .. " / " .. challenge.MaxCompleted)
				else
					wndStatus:SetText(challenge.TotalCompleted)
				end		
			end
		end
	end	

	--wrap it up
	local storedWindow = {}
	storedWindow.Window = wnd
	storedWindow.BtnToggle = wndBtnToggle
	storedWindow.Title = wndTitle
	storedWindow.Status = wndStatus
	storedWindow.ChallengeID = challengeID
	storedWindow.SortOrder = sortOrder or 0	
	return storedWindow
end

-- create an option list item then returns the window objects
function KiruChallenge:CreateOptionHeader(headerText, tooltipText, sortOrder, isWhiteList, isLockedList, isAutoStartList)
	-- load the window item for the list item
	local wnd = nil
	-- load the window item for the list item
	if isWhiteList then	wnd = Apollo.LoadForm(self.xmlDoc, "OptionListHeader", self.Option.ItemListWhite.Window, self) end
	if isLockedList then wnd = Apollo.LoadForm(self.xmlDoc, "OptionListHeader", self.Option.LockedList.Window, self) end
	if isAutoStartList then wnd = Apollo.LoadForm(self.xmlDoc, "OptionListHeader", self.Option.AutoStartList.Window, self) end
	if (not isLockedList) and (not isWhiteList) and (not isAutoStartList) then wnd = Apollo.LoadForm(self.xmlDoc, "OptionListHeader", self.Option.ItemList.Window, self) end

	wnd:SetTooltip(tooltipText)		
	wnd:SetText(headerText)
	local data = {SortOrder = sortOrder}
	wnd:SetData(data)
		--wrap it up
	local storedWindow = {}
	storedWindow.Window = wnd
	storedWindow.SortOrder = sortOrder	
	return storedWindow
end

-- create a list item for the PartySync window
function KiruChallenge:CreatePartySyncChallengeListItem(challengeID, challengeName, sortOrder)
	if challengeID == nil then return nil end
		
	-- load the window item for the list item
	local wnd = Apollo.LoadForm(self.xmlDoc, "PartySyncChallengeListItem", self.PartySync.ItemList.Window, self)	

	wnd:SetTooltip(challengeName)
	local data = {CID = challengeID}
	wnd:SetData(data)
	
	local wndTitle = wnd:FindChild("ChallengeName")
	local wndItemList = wnd:FindChild("ItemList")
	
	if wndTitle then wndTitle:SetText(challengeName) end

	--wrap it up
	local storedWindow = {}
	storedWindow.Window = wnd
	storedWindow.Title = wndTitle
	storedWindow.ItemList = wndItemList
	storedWindow.SortOrder = sortOrder
	return storedWindow
end

function KiruChallenge:CreatePartySyncPlayerListItem(parentWindow, playername, tier, status, time)
	if parentWindow == nil then return nil end

	-- load the window item for the list item
	local wnd = Apollo.LoadForm(self.xmlDoc, "PartySyncPlayerListItem", parentWindow, self)	
	wnd:SetTooltip(playername)

	local wndPlayer = wnd:FindChild("PlayerName")
	local wndStatus = wnd:FindChild("Status")
	local wndMedal = wnd:FindChild("Medal")
	local wndTime = wnd:FindChild("Time")
	local wndFinished = wnd:FindChild("Finished")
	
	if wndPlayer then wndPlayer:SetText(playername)	end
	if wndStatus then wndStatus:SetText(status)	end
	if wndTime then wndTime:SetText(time) end
	if wndMedal then self:UpdateMedalButton(wndMedal, tier) end

	--wrap it up
	local storedWindow = {}
	storedWindow.Window = wnd
	storedWindow.PlayerName = wndPlayer
	storedWindow.Status = wndStatus
	storedWindow.Time = wndTime
	storedWindow.Medal = wndMedal
	storedWindow.Finished = wndFinished
	return storedWindow	
end  

-- when a list item is selected
function KiruChallenge:OnListItemSelected(wndHandler, wndControl)
    -- make sure the wndControl is valid
    if wndHandler ~= wndControl then
        return
    end
        
	local challengeID = wndControl:GetData()
	self:ProcessSelected(challengeID)
end

-------------------------------------------------------------------------------------
-- Misc Challenge Functions
-------------------------------------------------------------------------------------

function KiruChallenge:TriggerStartChallenge(challengeID)
	if debugMode then self:PrintDebug("Triggering ID ".. challengeID) end
	self.StartingChallenge = challengeID
	ChallengesLib.ActivateChallenge(challengeID)
end

-------------------------------------------------------------------------------------
--Helper Functions
-------------------------------------------------------------------------------------

--silly LUA and its inability to count table entries using non-sequential integer keys
function KiruChallenge:GetTableLength(inputTable)
  	local tableCount = -1
	if inputTable then
		tableCount = 0
	  	for _ in pairs(inputTable) do tableCount = tableCount + 1 end
	end
  	return tableCount 
end

function KiruChallenge:PrintDebug(message)
	if debugMode then Print("KiruDebug: " .. message) end
end

--calculate the distance between two points in 3D space
function KiruChallenge:CalcDistance(posA, posB)
	local distance = -1
	--euclidian distance formula.. Square Root of (DeltaX squared + DeltaY squared + DeltaZ squared)
	distance = math.floor(math.sqrt((posA.x - posB.x)^2 + (posA.y - posB.y)^2 + (posA.z - posB.z)^2))
	return distance
end

function KiruChallenge:PrintVariable(input, label, indent, level, maxlevel)
	if label == nil then label = "nolabel" end
	if indent == nil then indent = 0 end
	if level == nil then level = 0 end
	if maxlevel == nil then maxlevel = 3 end
	
	local prepend = string.rep(" ",indent)
	local concat = prepend .. label .. " = "
	if input == nil then concat = concat .. "nil" return concat end
	local inputtype = string.lower(type(input))
	concat = concat .. "(" .. inputtype .. ") = "
	if inputtype == "table" then
		if level > maxlevel then
			concat = concat .. ">>> TOO DEEP"			
		else
			concat = concat .. "\n" .. prepend .. "{\n"
			for key, val in pairs(input) do
				concat = concat .. self:PrintVariable(val, key, indent + 4, level + 1, maxlevel)
			end
			concat = concat .. prepend .. "}"		
		end
	elseif inputtype == "function" then
		concat = concat .. "FUNCTION"
	elseif inputtype == "userdata" then
		concat = concat .. "USER DATA - Metatable Only"
		concat = concat .. "\n" .. prepend .. "{\n"
		concat = concat .. self:PrintVariable(getmetatable(input), "MetaTable", indent + 4, level + 1, maxlevel)
		concat = concat .. "}"
	else
		concat = concat .. tostring(input)
	end

	return concat .. "\n"
end

function KiruChallenge:Round(num, dp)
--rounds a number to specifed decimal places
	local mult = 10^(dp or 0)
	return math.floor(num * mult + 0.5) / mult
end

-----------------------------------------------------------------------------------------
--Table Sorting
-----------------------------------------------------------------------------------------
function KiruChallenge:GetKeysSortedByZoneThenName(inputTable)
	--sort challenges by zone then by challenge name, returns sorted keys

	--usage			
	--	for key, val in ipairs(keys) do
	--		whatever = tempTable[val]
	--	end
	
	local keys = {}
	for k in pairs(inputTable) do table.insert(keys, k) end
	table.sort(keys, function(a,b) 
		if inputTable[a].ZoneName == inputTable[b].ZoneName then 
			return inputTable[a].Name < inputTable[b].Name 
		else
			return inputTable[a].ZoneName < inputTable[b].ZoneName 
		end
	end)
	
	return keys
end

function KiruChallenge:GetKeysSortedByValue(inputTable)
	--sort an arbitrary [key]=value table on value
	--usage			
	--	for key, val in ipairs(keys) do
	--		whatever = tempTable[val]
	--	end
	
	local keys = {}
	for k in pairs(inputTable) do table.insert(keys, k) end
	table.sort(keys, function(a,b) 
		return inputTable[a] < inputTable[b] 
	end)
	
	return keys
end

function KiruChallenge:GetKeysSortedByAThenB(inputTable)
	--sort on A then B
	local keys = {}
	for k in pairs(inputTable) do table.insert(keys, k) end
	table.sort(keys, function(a,b) 
		if inputTable[a].A == inputTable[b].A then 
			return inputTable[a].B < inputTable[b].B 
		else
			return inputTable[a].A < inputTable[b].A 
		end
	end)	
	return keys
end


-----------------------------------------------------------------------------------------
-- Challenge Object and a function to print data about it
-----------------------------------------------------------------------------------------

function KiruChallenge:CreateChallengeObject(inputChallenge)
	--takes the challenge object returned by the API and
	--creates a more humanly-readible object for useful data that doesn't change
	--stores the original object inside .SourceChallenge for dynamic updates
	if inputChallenge == nil then return nil end

	local newChallengeObject = {}
	newChallengeObject.ID = inputChallenge:GetId()
	newChallengeObject.Name = inputChallenge:GetName()
	newChallengeObject.Description = inputChallenge:GetDescription()
	newChallengeObject.Type = inputChallenge:GetType()
	newChallengeObject.IsHouse = false
	local zoneInfo = inputChallenge:GetZoneInfo()
	if zoneInfo ~= nil then
		newChallengeObject.ZoneID = zoneInfo.idZone
		newChallengeObject.ZoneName = zoneInfo.strZoneName		
		--check if its a housing challenge
		if newChallengeObject.ZoneID == housingZone then newChallengeObject.IsHouse = true end
	end
	newChallengeObject.TotalCompleted = inputChallenge:GetCompletionCount()
	newChallengeObject.MaxCompleted = inputChallenge:GetCompletionTotal()
	newChallengeObject.MapLocation = inputChallenge:GetMapLocation()
	newChallengeObject.MapStartLocation = inputChallenge:GetMapStartLocation()
	newChallengeObject.StartLocationID = inputChallenge:GetStartLocationRestrictionId()
	newChallengeObject.ZoneRestriction = inputChallenge:GetZoneRestrictionInfo()
	newChallengeObject.SourceChallenge = inputChallenge

	return newChallengeObject
end

function KiruChallenge:GetChallengeDistance(challengeID,houseposition)
	--make sure the player has spawned...
	if GameLib:GetPlayerUnit() then
		local playerPos = GameLib:GetPlayerUnit():GetPosition()
		if houseposition then
			--using the distance of the house unit
			return self:CalcDistance(playerPos, houseposition)			
		end		
		--otherwise, using the stored data	
		local challenge = self.Data.AllChallenges[challengeID]	
		--if it has a start location, use that
		if challenge.MapStartLocation then
			return self:CalcDistance(playerPos, challenge.MapStartLocation)
		--else use its maplocation, which may be the center of a large area
		elseif challenge.MapLocation then
			return self:CalcDistance(playerPos, challenge.MapLocation)
		end
	end
	--no idea where the challenge is.. could be zone wide
	return -1
end
function KiruChallenge:PrintChallengeData(challengeID) 
	if challengeID and debugMode then
	
		local challenge = self.Data.AllChallenges[challengeID]
	
		local playerPos = GameLib:GetPlayerUnit():GetPosition()
		
		self:PrintDebug("For Challenge ID: " .. challenge.ID)
		self:PrintDebug(" Name = " .. challenge.Name)
		self:PrintDebug(" Description = " .. challenge.Description)
		self:PrintDebug(" TypeID = " .. challenge.Type)
		self:PrintDebug(" TypeName = " .. self.Data.ChallengeTypes[challenge.Type].Name)
		self:PrintDebug(" TimeTiered = " .. tostring(challenge.SourceChallenge:IsTimeTiered()))
		self:PrintDebug(self:PrintVariable(challenge.SourceChallenge:GetAllTierCounts(),"TierCounts"))
		self:PrintDebug(" ZoneID = " .. challenge.ZoneID)
		self:PrintDebug(" ZoneName = " .. challenge.ZoneName)
		if challenge.ZoneRestriction then
			self:PrintDebug(" ZoneRestriction = ")
			self:PrintDebug("    idSubZone = " .. challenge.ZoneRestriction.idSubZone)
			self:PrintDebug("    LocName = " .. challenge.ZoneRestriction.strLocationName)
			self:PrintDebug("    Name = " .. challenge.ZoneRestriction.strSubZoneName)
			self:PrintDebug("    IsPlayerThere = " .. tostring(GameLib.IsInWorldZone(challenge.ZoneRestriction.idSubZone)))
		else
			self:PrintDebug(" ZoneRestriction = nil")	
		end		
		self:PrintDebug(" TotalCompleted = " .. challenge.TotalCompleted)
		self:PrintDebug(" MaxCompleted = " .. challenge.MaxCompleted)
		if challenge.MapLocation then
			self:PrintDebug(" MapLoc = ")
			self:PrintDebug("    x = " .. challenge.MapLocation.x)
			self:PrintDebug("    y = " .. challenge.MapLocation.y)
			self:PrintDebug("    z = " .. challenge.MapLocation.z)
			self:PrintDebug(" Dist = " .. self:CalcDistance(playerPos, challenge.MapLocation))
		else
			self:PrintDebug(" MapLoc = nil")	
		end		
		if challenge.MapStartLocation then
			self:PrintDebug(" MapStartLoc = ")
			self:PrintDebug("    x = " .. challenge.MapStartLocation.x)
			self:PrintDebug("    y = " .. challenge.MapStartLocation.y)
			self:PrintDebug("    z = " .. challenge.MapStartLocation.z)
			self:PrintDebug(" Dist = " .. self:CalcDistance(playerPos, challenge.MapStartLocation))
		else
			self:PrintDebug(" MapStartLoc = nil")	
		end		
		self:PrintDebug(" StartLocationID = " .. challenge.StartLocationID)
		self:PrintDebug("    IsPlayerThere = " .. tostring(GameLib.IsInLocation(challenge.StartLocationID)))
		self:PrintDebug(" .......")
		local MapLocTwo = challenge.SourceChallenge:GetMapLocation()
		if MapLocTwo then
			self:PrintDebug(" MapLoc2 = ")
			self:PrintDebug("    x = " .. MapLocTwo.x)
			self:PrintDebug("    y = " .. MapLocTwo.y)
			self:PrintDebug("    z = " .. MapLocTwo.z)
			self:PrintDebug(" Dist = " .. self:CalcDistance(playerPos, MapLocTwo))
		else
			self:PrintDebug(" MapLoc2 = nil")	
		end	
		local MapStartLocTwo = challenge.SourceChallenge:GetMapStartLocation()		
		if MapStartLocTwo then
			self:PrintDebug(" MapStartLoc2 = ")
			self:PrintDebug("    x = " .. MapStartLocTwo.x)
			self:PrintDebug("    y = " .. MapStartLocTwo.y)
			self:PrintDebug("    z = " .. MapStartLocTwo.z)
			self:PrintDebug(" Dist = " .. self:CalcDistance(playerPos, MapStartLocTwo))
		else
			self:PrintDebug(" MapStartLoc2 = nil")	
		end		
	end
end

---------------------------------------------------------------------------------------------------
-- Button Functions
---------------------------------------------------------------------------------------------------

function KiruChallenge:AddToMainChallengeList(challengeID)
	--if it's not in the current list, add it
	if self.Main.ItemList.Items[challengeID] == nil then
		self.Main.ItemList.Items[challengeID] = self:CreateChallengeListItem(challengeID)
		--add it to the map
		--remove it first if it was locked
		self:RemoveChallengeMap(challengeID)
		self:AddChallengeMap(self.Data.AllChallenges[challengeID])
	end
end

function KiruChallenge:RemoveFromMainChallengeList(challengeID)
	if self.Main.ItemList.Items and self.Main.ItemList.Items[challengeID] ~= nil then 
		--delete the window item first
		self.Main.ItemList.Items[challengeID].Window:Destroy()
		self.Main.ItemList.Items[challengeID] = nil	
		--remove it from the map
		self:RemoveChallengeMap(challengeID)	
	end
end

function KiruChallenge:Button_Toggle(wndHandler, wndControl, eMouseButton )
	local checked = wndControl:IsChecked()
	local data = wndControl:GetParent():GetData()
	local challengeID = data.ChallengeID
	local isWhiteList = data.IsWhiteList
	local isAutoStart = data.IsAutoStart
	
	if isWhiteList then
		if checked then
			--add to white list and to the challenge list if it's not already there
			self.Data.Preferences.WhiteList[challengeID] = 1
			self:AddToMainChallengeList(challengeID)
			--remove from ignore list if it's there
			if self.Data.Preferences.IgnoreList[challengeID] ~= nil then
				self.Data.Preferences.IgnoreList[challengeID] = nil
				if self.Option.ItemList.Items and self.Option.ItemList.Items[challengeID] ~= nil then
					self.Option.ItemList.Items[challengeID].BtnToggle:SetCheck(false)
				end
			end			
		else
			self.Data.Preferences.WhiteList[challengeID] = nil
			--remove from challenge list if it doesn't belong there
			if self:ShouldDisplayChallenge(challengeID) ~= 1 then self:RemoveFromMainChallengeList(challengeID) end
		end
	else
		if isAutoStart then
			if self.Data.Preferences.AutoStart[challengeID] == nil then	self.Data.Preferences.AutoStart[challengeID] = {} end
			local distance = self.Data.Preferences.AutoStart[challengeID].D or 10
			if checked then
				self.Data.Preferences.AutoStart[challengeID] = { O = 1, D = distance }
			else
				self.Data.Preferences.AutoStart[challengeID] = { O = 0, D = distance }
			end
		else
			if checked then
				--add to ignore list
				self.Data.Preferences.IgnoreList[challengeID] = 1
				--if it's in the current list, remove it - unless it has a reward
				if self:ShouldDisplayChallenge(challengeID) ~= 2 then self:RemoveFromMainChallengeList(challengeID) end
				--remove from white list if it's there
				if self.Data.Preferences.WhiteList[challengeID] ~= nil then
					self.Data.Preferences.WhiteList[challengeID] = nil
					if self.Option.ItemListWhite.Items and self.Option.ItemListWhite.Items[challengeID] ~= nil then
						self.Option.ItemListWhite.Items[challengeID].BtnToggle:SetCheck(false)
					end
				end			
			else
				self.Data.Preferences.IgnoreList[challengeID] = nil
				--if it's a challenge that is in the area, show it
				if self:ShouldDisplayChallenge(challengeID) == 1 then self:AddToMainChallengeList(challengeID) end
			end						
		end
	end
end

function KiruChallenge:Button_ShowOption()
	if self.Option == nil or self.Option.Window == nil then
		self:CreateOptionWindow()
	else
		self:Button_HideOption()
	end	
end

function KiruChallenge:Button_HideOption()
	--overwrite the option window position then destroy
	local l,t,r,b = self.Option.Window:GetAnchorOffsets()
	if self.Data.Preferences == nil then self.Data.Preferences = {} end
	if self.Data.Preferences.FromRestore == nil then self.Data.Preferences.FromRestore = {} end
	self.Data.Preferences.FromRestore.OptionWindowPos = { l = l, t = t, r = r, b = b}
	self.Option.Window:Destroy()
	self.Option = {}
end

function KiruChallenge:Button_HideMain()
	self.ClosedOnPurpose = true
	self.Main.Window:Show(false)
	self.CombatViewState = false
end

function KiruChallenge:Button_KeybindCapture( wndHandler, wndControl, eMouseButton )
	self.CaptureKeybind = true
	wndControl:Enable(false)
	wndControl:SetText("Press A Key")
end

function KiruChallenge:Button_CaptureClear( wndHandler, wndControl, eMouseButton )
	self.Keybind = nil
	self.KeybindChallenge = nil
	self.Option.Window:FindChild("KeybindEditbox"):SetText("")
	Print("KiruChallenge: Keybind Cleared")
end

function KiruChallenge:Button_KeybindNextCapture( wndHandler, wndControl, eMouseButton )
	self.CaptureNextKeybind = true
	wndControl:Enable(false)
	wndControl:SetText("Press A Key")
end

function KiruChallenge:Button_NextCaptureClear( wndHandler, wndControl, eMouseButton )
	self.KeybindNext = nil
	self.Option.Window:FindChild("KeybindNextEditbox"):SetText("")
	Print("KiruChallenge: Keybind for Toggle Cleared")
end


function KiruChallenge:Button_HideNoChallenge( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.HideNoChallenges = wndControl:IsChecked()
end

function KiruChallenge:Button_PartySpam( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.PartySpam = wndControl:IsChecked()
	if self.Data.Preferences.PartySpam then
		Print("KiruChallenge: Spam ON. Will send Start and Finish messages to Party.")
		if self.Data.Preferences.PartySpamLoot then Print("KiruChallenge: Loot messages too!") end
		--enable loot spam
		self.Option.Window:FindChild("Button_PartySpamLoot"):Enable(true)
	else
		Print("KiruChallenge: Spam OFF. No messages will be sent to Party.")
		--disable loot spam
		self.Option.Window:FindChild("Button_PartySpamLoot"):Enable(false)
	end
end

function KiruChallenge:Button_PartySync( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.PartySync = wndControl:IsChecked()
	
	if self.Data.Preferences.PartySync then
		Print("KiruChallenge: Party Sync ON")
		self.Option.RadioSyncBottom:Enable(true)
		self.Option.RadioSyncHide:Enable(true)
		self.Option.RadioSyncTop:Enable(true)
		self.Option.RadioSyncFloat:Enable(true)
	else
		Print("KiruChallenge: Party Sync OFF")	
		self.Option.RadioSyncBottom:Enable(false)
		self.Option.RadioSyncHide:Enable(false)
		self.Option.RadioSyncTop:Enable(false)
		self.Option.RadioSyncFloat:Enable(false)
	end
	
	if self.Data.Preferences.PartySync and GroupLib.InGroup() then
		self:StartPartySync()	
	else
		self:StopPartySync()
	end
end

function KiruChallenge:Button_ClearSync( wndHandler, wndControl, eMouseButton )
	self:ClearPartySyncList()
end

---------------------------------------------------------------------------------------------------
-- KiruChallengeOptionForm Functions
---------------------------------------------------------------------------------------------------

function KiruChallenge:OnSystemKeyDown(iKey)
	--rudimentary and temporary keybinding
	if self.CaptureKeybind then
		if self.Option == nil or self.Option.Window == nil then
			self.CaptureKeybind = false
			return
		end
		local keytext = tblSpecialKeys[iKey] or string.char(iKey)
		if keytext == "SHF" then return end	
		if self.KeybindNext and iKey == self.KeybindNext then return end
		self.Keybind = iKey
		local button = self.Option.Window:FindChild("Button_Capture")
		local textbox = self.Option.Window:FindChild("KeybindEditbox")
		button:Enable(true)
		button:SetText("Capture")
		textbox:SetText(keytext)
		Print("KiruChallenge: Keybind Set to " .. keytext)
		self.CaptureKeybind = false
		return	
	end
	if self.CaptureNextKeybind then
		if self.Option == nil or self.Option.Window == nil then
			self.CaptureNextKeybind = false
			return
		end
		local keytext = tblSpecialKeys[iKey] or string.char(iKey)	
		if keytext == "SHF" then return end	
		if self.Keybind and iKey == self.Keybind then return end
		self.KeybindNext = iKey
		local button = self.Option.Window:FindChild("Button_CaptureNext")
		local textbox = self.Option.Window:FindChild("KeybindNextEditbox")
		button:Enable(true)
		button:SetText("Capture")
		textbox:SetText(keytext)
		Print("KiruChallenge: Keybind for Toggle Set to " .. keytext)
		self.CaptureNextKeybind = false
		return	
	end
	if self.Keybind and iKey == self.Keybind then
		--activate the topmost "OK" challenge
		if self.KeybindChallenge ~= nil then
			self:ProcessSelected(self.KeybindChallenge)
		end
	end
	if self.KeybindNext and iKey == self.KeybindNext then
		if Apollo.IsShiftKeyDown() then
			self.KeybindShiftDown = self.KeybindShiftDown - 1		
		else
			self.KeybindShiftDown = self.KeybindShiftDown + 1		
		end
		self.KeybindShiftOnce = true
	end
end

function KiruChallenge:ProcessSelected(challengeID)
	if debugMode then self:PrintDebug( "Item " ..  challengeID .. " is selected.") end
	
	--statuses:  OK, Reward, Area, NoneLeft, Cooling, Active, Missing
	--decide what to do with the click, based on the current status
	--either:  claim reward
	--         show an arrow
	--         do nothing
	--         start the challenge
	
	--get current status
	local curStatus = self:GetChallengeStatus(challengeID)
	if curStatus ~= nil and curStatus.Status ~= nil then
		if debugMode then self:PrintDebug("Selected... status = " .. curStatus.Status) end
		--if it's a reward, all good
		if curStatus.Status == Status_Reward then
			Event_FireGenericEvent("ChallengeRewardShow", challengeID)
			return
		end
		--if it's OK but we're in a raid, can't start it.
		if curStatus.Status == Status_OK then
			if GroupLib.InRaid() then 
				Print("KiruChallenge: You cannot start a challenge while in a Raid Party.  Sorry.")
			else
				self:TriggerStartChallenge(challengeID)
			end
			return
		end
		if curStatus.Status == Status_Area then
			ChallengesLib.ShowHintArrow(challengeID)
		end
	end
end

function KiruChallenge:IgnoreList_Hide( wndHandler, wndControl )
	if self.Option == nil then return end
	if self.Option.ItemList == nil then return end
	self:SyncScrollBars(self.Option.ItemList.Window:GetVScrollPos())
end

function KiruChallenge:WhiteList_Hide( wndHandler, wndControl )
	if self.Option == nil then return end
	if self.Option.ItemListWhite == nil then return end
	self:SyncScrollBars(self.Option.ItemListWhite.Window:GetVScrollPos())
end

function KiruChallenge:AutoStart_Hide( wndHandler, wndControl )
	if self.Option == nil then return end
	if self.Option.AutoStartList == nil then return end
	self:SyncScrollBars(self.Option.AutoStartList.Window:GetVScrollPos())
end

function KiruChallenge:SyncScrollBars(position)
	if self.Option == nil then return end
	if self.Option.ItemListWhite.Window then
		self.Option.ItemListWhite.Window:SetVScrollPos(position)
	end
	if self.Option.ItemList.Window then
		self.Option.ItemList.Window:SetVScrollPos(position)
	end	
	if self.Option.AutoStartList.Window then
		self.Option.AutoStartList.Window:SetVScrollPos(position)
	end
end

function KiruChallenge:Slider_Opacity( wndHandler, wndControl, fNewValue, fOldValue )
	local newValue = math.floor(fNewValue)
	local oldValue = math.floor(fOldValue)
	if newValue == oldValue  then return end
	local newValuePerc = self:Round((newValue*100) / 255, 0)
	local newValueHex = string.format('%x',newValue)
	if newValuePerc < 16 then newValueHex = "0"..newValueHex end
	self.Data.Preferences.Opacity = newValuePerc
	self.Option.OpacityLabel:SetText(newValuePerc.."%")
	self.Main.BGTransparency:SetBGColor(newValueHex.."FFFFFF")
	if self.PartySync.IsShown and self.PartySync.IsFloat then
		self.PartySync.Float:FindChild("TransparentBG"):SetBGColor(newValueHex .."FFFFFF")	
	end
end

function KiruChallenge:Button_DisplayScrollDown( wndHandler, wndControl, eMouseButton )
	if self.Data.Preferences.ShowScroll > 1 then
		self.Data.Preferences.ShowScroll = self.Data.Preferences.ShowScroll - 1
		self.Option.Window:FindChild("ShowScrollEdit"):SetText(self.Data.Preferences.ShowScroll)
		self:ResizeMainWindow()
	end
end

function KiruChallenge:Button_DisplayScrollUp( wndHandler, wndControl, eMouseButton )
	if self.Data.Preferences.ShowScroll < 50 then
		self.Data.Preferences.ShowScroll = self.Data.Preferences.ShowScroll + 1
		self.Option.Window:FindChild("ShowScrollEdit"):SetText(self.Data.Preferences.ShowScroll)
		self:ResizeMainWindow()
	end
end

function KiruChallenge:Radio_SyncPosition( wndHandler, wndControl, eMouseButton )
	local newPosition = wndControl:GetData().Pos
	local oldPosition = self.PartySync.Position
	if oldPosition == newPosition then return end

	self.PartySync.Position = newPosition
	
	if newPosition == "H" and self.PartySync.IsShown then
		self:DestroyPartySyncWindow()
		return
	end

	if self.PartySync.IsShown then
		if self.PartySync.IsFloat and (newPosition == "B" or newPosition == "T") then
			self:DestroyPartySyncWindow()
			self:CreatePartySyncWindow(true)
		end
		if not self.PartySync.IsFloat and newPosition == "F" then
			self:DestroyPartySyncWindow()
			self:CreatePartySyncWindow(true)		
		end
		if not self.PartySync.IsFloat and (newPosition == "B" or newPosition == "T") then
			self:ArrangeMainWindow()
		end	
	else
		self:UpdateSync()
	end
end

function KiruChallenge:Button_AutoHideScrollUp( wndHandler, wndControl, eMouseButton )
	if self.Data.Preferences.SyncAutoHideTime < 300 then
		self.Data.Preferences.SyncAutoHideTime = self.Data.Preferences.SyncAutoHideTime + 1
		self.Option.Window:FindChild("AutoHideEdit"):SetText(self.Data.Preferences.SyncAutoHideTime)
	end
end

function KiruChallenge:Button_AutoHideScrollDown( wndHandler, wndControl, eMouseButton )
	if self.Data.Preferences.SyncAutoHideTime > 10 then
		self.Data.Preferences.SyncAutoHideTime = self.Data.Preferences.SyncAutoHideTime - 1
		self.Option.Window:FindChild("AutoHideEdit"):SetText(self.Data.Preferences.SyncAutoHideTime)
	end
end

function KiruChallenge:Button_PartySyncAutoHide( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.SyncAutoHide = wndControl:IsChecked()
	if not self.Data.Preferences.SyncAutoHide then
		if self.PartySync.HideButton ~= nil then self.PartySync.HideButton:SetText("Hide") end
	end
end

function KiruChallenge:Button_PartySpamLoot( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.PartySpamLoot = wndControl:IsChecked()
	if self.Data.Preferences.PartySpam then
		Print("KiruChallenge: Spam ON. Will send Start and Finish messages to Party.")
		if self.Data.Preferences.PartySpamLoot then 
			Print("KiruChallenge: Loot messages too!")
		else
			Print("KiruChallenge: Loot messages will no longer be included in Party Spam.")	
		end
	end
end

function KiruChallenge:Button_ShowMap( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.ShowMap = wndControl:IsChecked()
	self:ShowMapChallenges(wndControl:IsChecked())
end

function KiruChallenge:Button_ShowNub( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.ShowKeybindNub = wndControl:IsChecked()
end

function KiruChallenge:Button_AutoReward( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.AutoReward = wndControl:IsChecked()
end

function KiruChallenge:Button_ReadyDistance( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.ReadyDistance = wndControl:IsChecked()
end

function KiruChallenge:Slider_Distance( wndHandler, wndControl, fNewValue, fOldValue )
	local newValue = math.floor(fNewValue)
	local oldValue = math.floor(fOldValue)
	if newValue == oldValue  then return end
	self.Data.Preferences.FilterDistance = newValue
	if newValue == 2000 then
		self.Option.Window:FindChild("DistanceLabel"):SetText("Show All")
	else
		self.Option.Window:FindChild("DistanceLabel"):SetText(newValue.."m")	
	end
end

function KiruChallenge:Button_ShowCooldown( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.ShowCooldown = wndControl:IsChecked()
end

function KiruChallenge:Button_HideCombat( wndHandler, wndControl, eMouseButton )
	self.Data.Preferences.HideCombat = wndControl:IsChecked()
	self:EnteredCombatSpam(self.Data.Preferences.HideCombat)
end

function KiruChallenge:OnEnteredCombat(unit, inCombat)
	if self.Data.Preferences.HideCombat == false then
		--turn it off
		self:EnteredCombatSpam(false)
		return
	end
	if unit == nil then return end
	if self.PlayerName == nil then return end
	if unit:GetName() ~= self.PlayerName then return end

	self.InCombat = inCombat

	if inCombat then
		self.Main.Window:Show(false,true)
		if self.PartySync.IsFloat then
			self.PartySync.Float:Show(false,true)
		end
	else
		if self.CombatViewState == nil then self.CombatViewState = true end
		if self.CombatSyncViewState == nil then self.CombatSyncViewState = true end
		self.Main.Window:Show(self.CombatViewState,true)
		if self.PartySync.IsFloat then
			self.PartySync.Float:Show(self.CombatSyncViewState,true)
		end
	end
end

function KiruChallenge:EnteredCombatSpam(enable)
	if enable then
		Apollo.RegisterEventHandler("UnitEnteredCombat", "OnEnteredCombat", self)
	else
		Apollo.RemoveEventHandler("UnitEnteredCombat", self)
	end	
end	

function KiruChallenge:Radio_Shared( wndHandler, wndControl )
	local pref = wndControl:GetData()
	GameLib.SetSharedChallengePreference(pref)
	g_InterfaceOptions.Carbine.eShareChallengePreference = pref
	Event_FireGenericEvent("ChallengeLog_UpdateShareChallengePreference", pref)
	if pref == 1 then Print("Should Prompt to start a challenge that a group member initiates.") end
	if pref == 0 then Print("Should automatically accept all challenges that a group member initiates.") end
	if pref == 2 then Print("Should automatically reject all challenges that a group member initiates.") end
end

---------------------------------------------------------------------------------------------------
-- Party Sync stuff
---------------------------------------------------------------------------------------------------
--events
function KiruChallenge:OnGroupJoin()
	if self.Data.Preferences.PartySync then self:StartPartySync() end
end

function KiruChallenge:OnGroupLeft()
	self:StopPartySync()
end

function KiruChallenge:OnGroupRemove()
	--if we had sync, check to see if we still have sync
	--if a test message fails, we no longer do.
	--if someone else joins the party with this addon, theirs will initiate sync again
	if self.HasPartySync then
		local testmsg = { Text = "PING" }
		local testsync = self.ChatChannel:SendMessage(testmsg)
		if not testsync then
			self.HasPartySync = false
			self:ShowSyncBar(false)
		end
	end
end

function KiruChallenge:OnLeaderChange(leaderName)
	if self.Data.Preferences.PartySync then self:StartPartySync() end
end

function KiruChallenge:StartPartySync()	
	self:StopPartySync(true)
	self.StartSync = 0
	self.StopSync = 0

	self.PartyLeader = self:GetPartyLeader()
	if self.PartyLeader == nil then return end

	local channelName = "ICCKC" .. self.PartyLeader
	self:PrintDebug("Joining Channel " .. channelName)
	self.ChatChannel = ICCommLib.JoinChannel(channelName, "OnChannelChat", self)
	if self.ChatChannel ~= nil then self:PrintDebug("Channel Join Success") else self:PrintDebug("Channel Join Failed") end
	if self.ChatChannel == nil then return end
	
	--wait 4 seconds before sending sync request.  this will be handled in the timer
	--there's apparently an issue with the channel taking a few seconds before accepting messages
	if debugMode then self:PrintDebug("Starting Sync in 4 seconds.") end
	self.StartSync = 4
end

function KiruChallenge:StopPartySync(isStarting)
	--if we're in a group and we aren't starting the sync,
	--send out a message that we've turned it off.
	if GroupLib.InGroup() and not isStarting and self.HasPartySync then
		local stopmsg = { Text = "STOP" }
		self.ChatChannel:SendMessage(stopmsg)
	end
	self.ChatChannel = nil
	self.PartyLeader = nil
	self.HasPartySync = false
	self:ShowSyncBar(false)
end

function KiruChallenge:GetPartyLeader()
	local groupCount = GroupLib.GetMemberCount()
	if groupCount > 0 then
		for idx = 1, groupCount do
			local member = GroupLib.GetGroupMember(idx)
			if member ~= nil and member.bIsLeader then
				return member.strCharacterName
			end
		end	
	end
	return nil
end

function KiruChallenge:OnChannelChat(channel, tMsg)
	--there's no way to actually leave the channel once we join it.  has to be garbage collected by the server
	--so we need to make sure we're wanting this message before doing anything with it
	if not self.Data.Preferences.PartySync or self.ChatChannel == nil then
		if not partyOfOne then return end
	end
	--final check to make sure we aren't receiving messages from a channel we aren't part of.
	--or if we're sending ourself an update
	if channel ~= "self" and self.ChatChannel ~= channel then return end
	
	if debugMode then self:PrintDebug("Addon Msg Recieved:") end
	if tMsg == nil then self:PrintDebug("Empty Message") return end
	--specifying debugMode here so it doesn't have to loop through the variable just to not print it later
	if debugMode then self:PrintDebug(self:PrintVariable(tMsg, "tMsg", 0, 0, 3)) end
	
	--receiving a sync request
	if tMsg.Text == "PING" then
		local msg = {}
		msg.Text = "PONG"		
		local didSend = self.ChatChannel:SendMessage(msg)
		if debugMode then self:PrintDebug("Received Party Sync Request and replied " .. tostring(didSend)) end
		self.HasPartySync = true
		self:ShowSyncBar(true)
		self:UpdateSync("ON")
		self.StopSync = 0
		self.StartSync = 0
		self.WaitSync = 0
		if debugMode then self:PrintDebug("Party Sync Active") end
	end
	--receiving a sync answer to our request
	if tMsg.Text == "PONG" then
		if debugMode then self:PrintDebug("Received Sync Reply") end
		--someone in the party is using the addon.  turn sync on
		self.HasPartySync = true
		self:ShowSyncBar(true)
		self:UpdateSync("ON")
		self.StopSync = 0
		self.StartSync = 0
		self.WaitSync = 0
		if debugMode then self:PrintDebug("Party Sync Active") end
	end
	--receiving a stop message
	--turn off HasGroupSync in 2 seconds if no reply is received from a ping
	if tMsg.Text == "STOP" then
		local msg = {}
		msg.Text = "PING"		
		local didSend = self.ChatChannel:SendMessage(msg)
		if debugMode then self:PrintDebug("Received a Stop sync message.. sending Ping and waiting 3 seconds " .. tostring(didSend)) end
		self.StopSync = 3	
	end
	--receiving an update
	if tMsg.Text == "UPDATE" or tMsg.Text == "FINAL" then
		if not self.PartySync.IsShown and not self.ClosedSyncOnPurpose and self.PartySync.Position ~= "H" then 
			self:ShowSync(true)
		end	

		if self.PartySync.IsShown and self.Data.Preferences.SyncAutoHide then
			self.HideSyncAuto = self.Data.Preferences.SyncAutoHideTime
		end
			
		local challengeID = tMsg.CID
		--check to see if the challenge is in the sync window.  if not, create it
		if self.PartySync.ItemList.Items == nil then self.PartySync.ItemList.Items = {} end
		if self.PartySync.ItemList.Items[challengeID] == nil then
			--first re-sort the others by setting their order to +1, newest on top
			if self:GetTableLength(self.PartySync.ItemList.Items) > 0 then
				for key, listItem in pairs(self.PartySync.ItemList.Items) do
					listItem.SortOrder = listItem.SortOrder + 1
				end
			end
			self.PartySync.ItemList.Items[challengeID] = self:CreatePartySyncChallengeListItem(challengeID, tMsg.CName, 1)
			self.PartySync.ItemList.Window:ArrangeChildrenVert(0, function(a,b)
					local idA = a:GetData().CID
					local idB = b:GetData().CID			
					return self.PartySync.ItemList.Items[idA].SortOrder < self.PartySync.ItemList.Items[idB].SortOrder
				end
			)					
		end
		local parentChallengeWindow = self.PartySync.ItemList.Items[challengeID]
		--initialize the player list if it's not already there
		if parentChallengeWindow.PlayerList == nil then parentChallengeWindow.PlayerList = {} end
		
		local playerName = tMsg.Player
		local currentCount = tMsg.Current
		local isTimed = tMsg.IsTime
		local tierNumber = tMsg.DTier
		local timeText = tMsg.Time
		if (not tMsg.HasGold) and (tMsg.DTier == 1) then tierNumber = 3 end	
		
		if isTimed then currentCount = "Timed" end
		
		local playerWindow = nil

		--check to see if the player is in the challenge window.  if not, create it		
		if parentChallengeWindow.PlayerList[tMsg.Player] == nil then
			parentChallengeWindow.PlayerList[tMsg.Player] = self:CreatePartySyncPlayerListItem(parentChallengeWindow.ItemList, playerName, tierNumber, currentCount, timeText)
			parentChallengeWindow.ItemList:ArrangeChildrenVert()
			playerWindow = parentChallengeWindow.PlayerList[tMsg.Player]
		else
			--otherwise update the player information
			playerWindow = parentChallengeWindow.PlayerList[tMsg.Player]
			playerWindow.Status:SetText(currentCount)			
			playerWindow.Time:SetText(timeText)
			--.Medal
			self:UpdateMedalButton(playerWindow.Medal,tierNumber)
		end

		--.Finished?		
		if tMsg.Text == "FINAL" then
			if tMsg.Status == 1 then
				self:ShowFinalButton(true, playerWindow, true)							
			else
				self:ShowFinalButton(true, playerWindow, false)			
			end
		else
			self:ShowFinalButton(false, playerWindow)		
		end
	end
end

function KiruChallenge:ShowFinalButton(show, playerWin, isComplete)
	local l, t, r, b = playerWin.PlayerName:GetAnchorOffsets()
	if show then
		--move playername over
		l = 18
		--pick a sprite
		if isComplete then
			playerWin.Finished:SetSprite("CRB_Basekit:kitIcon_CheckmarkGreenFrame")
		else
			playerWin.Finished:SetSprite("ClientSprites:LootCloseBox_Holo")
		end		
	else
		--move playername back
		l = 5
	end
	playerWin.PlayerName:SetAnchorOffsets(l,t,r,b)
	playerWin.Finished:Show(show,true)	
end

function KiruChallenge:UpdateMedalButton(medalButton, tier)
	local tierData = self.Data.Tiers[tier]
	local toolTip = tierData.Name or "N/A"
	local BG = tierData.BGColor or "FF777777"
	local TC = tierData.TextColor or "FFFFFFFF"
	local text = tierData.Abbrev or ""
	
	medalButton:SetBGColor(BG)
	medalButton:SetTextColor(TC)
	medalButton:SetText(text)
	medalButton:SetTooltip(toolTip)
end

function KiruChallenge:ToggleSync(onPurpose)
	if self.PartySync.Position == "H" then return end
	if self.PartySync.IsShown then
		self:HideSync()
	else
		self:ShowSync()
	end
end

function KiruChallenge:ShowSync(autoHide)
	if self.PartySync.Position == "H" then return end
	if self.PartySync.Window == nil then 
		self.ClosedSyncOnPurpose = false
		self:CreatePartySyncWindow(true)
		self:UpdateSync()
		return
	end
	if self.PartySync.HideButton ~= nil then self.PartySync.HideButton:SetText("Hide") end
	self.PartySync.IsShown = true
	if self.PartySync.IsFloat then
		if self.Data.Preferences.HideCombat == true and self.InCombat == true then
			self.CombatSyncViewState = true
		else
			self.PartySync.Float:Show(true,true)
		end
		self.CombatSyncViewState = true
	else
		self.PartySync.Dock:Show(true,true)
		self:ArrangeMainWindow()
	end	
	self.ClosedSyncOnPurpose = false
	self:UpdateSync()
end

function KiruChallenge:HideSync()
	self.PartySync.IsShown = false
	if self.PartySync.IsFloat then
		self.PartySync.Float:Show(false,true)	
		self.CombatSyncViewState = false
	else
		self.PartySync.Dock:Show(false,true)
		self:ArrangeMainWindow()
	end
end

function KiruChallenge:CloseSync()
	self.ClosedSyncOnPurpose = true	
	self:DestroyPartySyncWindow()
	self:UpdateSync()
end

function KiruChallenge:CreatePartySyncWindow(showAfterCreate)
	if self.PartySync.IsShown then return end
	if self.PartySync.Position == "F" then
		self.PartySync.Window = Apollo.LoadForm(self.xmlDoc, "PartySyncGuts", self.PartySync.Float, self)
		self.PartySync.IsFloat = true
		if self.Data.Preferences.FromRestore ~= nil then
			local anchor = self.Data.Preferences.FromRestore.PartySyncWindowPos
			if anchor then self.PartySync.Float:SetAnchorOffsets(anchor.l, anchor.t, anchor.r, anchor.b) end
		end
		self.PartySync.Float:SetSizingMinimum(280,syncsize)
		
		--opacity
		local opacityPre = (self.Data.Preferences.Opacity / 100) * 255
		local opacityHex = string.format('%x',opacityPre)
		if opacityPre < 16 then opacityHex = "0"..opacityHex end
		self.PartySync.Float:FindChild("TransparentBG"):SetBGColor(opacityHex .."FFFFFF")			
		if self.Data.Preferences.HideCombat == true and self.InCombat == true then
			self.CombatSyncViewState = showAfterCreate
		else
			self.PartySync.Float:Show(showAfterCreate,true)
		end
		self.CombatSyncViewState = showAfterCreate				
	else
		self.PartySync.Window = Apollo.LoadForm(self.xmlDoc, "PartySyncGuts", self.PartySync.Dock, self)
		self.PartySync.IsFloat = false
		self.PartySync.Dock:Show(showAfterCreate, true)
	end
	self.PartySync.IsShown = showAfterCreate
	
	self.PartySync.ItemList = {}
	self.PartySync.ItemList.Items = {}	
	self.PartySync.ItemList.Window = self.PartySync.Window:FindChild("SyncItemList")
	
	self.PartySync.HideButton = self.PartySync.Window:FindChild("Button_Hide")
	
	if showAfterCreate and not self.PartySync.IsFloat then self:ArrangeMainWindow() end
	self:UpdateSync()
end

function KiruChallenge:DestroyPartySyncWindow()
	if self.PartySync.Window == nil then return end
	if self.PartySync.IsFloat then
		local l,t,r,b = self.PartySync.Float:GetAnchorOffsets()
		if self.Data.Preferences == nil then self.Data.Preferences = {} end
		if self.Data.Preferences.FromRestore == nil then self.Data.Preferences.FromRestore = {} end
		self.Data.Preferences.FromRestore.PartySyncWindowPos = { l = l, t = t, r = r, b = b}	
	end
	self.PartySync.Window:Destroy()
	self.PartySync.IsShown = false
	self.PartySync.Window = nil	
	self.PartySync.HideButton = nil
	if self.PartySync.IsFloat then
		self.PartySync.Float:Show(false,true)
		self.CombatSyncViewState = false		
	else
		self.PartySync.Dock:Show(false,true)
		self:ArrangeMainWindow()
	end
	self:UpdateSync()
end

function KiruChallenge:ArrangeMainWindow()
	local mainSection = self.Main.Window:FindChild("MainSection")
	local ml,mt,mr,mb = mainSection:GetAnchorOffsets()
	local mh = mb - mt

	if not self.PartySync.IsShown or self.PartySync.IsFloat or self.PartySync.Position == "H" then
		--make sure the main section is at the top and sync is hidden
		mainSection:SetAnchorOffsets(ml,0,mr,mh)
		self.PartySync.Dock:Show(false,true)
	else
		local sl,st,sr,sb = self.PartySync.Dock:GetAnchorOffsets()
		local sh = sb - st
		
		--determine where to place the sync window
		if self.PartySync.Position == "T" then
			--partysync goes first
			self.PartySync.Dock:SetAnchorOffsets(sl,0,sr,sh)
			mainSection:SetAnchorOffsets(ml,sh,mr,sh + mh)
		else
			--mainsection goes first
			mainSection:SetAnchorOffsets(ml,0,mr,mh)
			self.PartySync.Dock:SetAnchorOffsets(sl,mh,sr,mh + sh)		
		end
	end
	self:ResizeMainWindow()
end

function KiruChallenge:ResizeMainWindow()
	local mainBase = 20 --title 20 + 4 at the bottom
	local padding = 8
	local challengeList = (self.Data.Preferences.ShowScroll * 20) +2
	local syncDock = 0

	--set the main section height based on the shown challenges
	local mainSection = self.Main.Window:FindChild("MainSection")
	local l,t,r,b = mainSection:GetAnchorOffsets()	
	b = t + mainBase + challengeList
	mainSection:SetAnchorOffsets(l,t,r,b)
	
	--if sync is open and not floating, dock it under the new mainSection bottom
	if self.PartySync.IsShown and not self.PartySync.IsFloat then
		syncDock = syncsize
		if self.PartySync.Position == "B" then
			local sl,st,sr,sb = self.PartySync.Dock:GetAnchorOffsets()
			self.PartySync.Dock:SetAnchorOffsets(sl, b, sr, b + syncDock)	
		end
	end
		
	--adjust the overall window
	l,t,r,b = self.Main.Window:GetAnchorOffsets()
	
	if self.PartySync.IsShown and (not self.PartySync.IsFloat) and (self.PartySync.Position ~= "H") then
		--is positioned at the top and we haven't shifted things up yet, do so.
		if self.PartySync.Position == "T" and self.Main.OutPosition == 0 then
			--shift it up
			t = t - syncsize 
			b = b - syncsize
			self.Main.OutPosition = 1		
		end
		--is positioned at the bottom and we shifted things up once, move it back down.
		if self.PartySync.Position == "B" and self.Main.OutPosition == 1 then
			--shift it down
			t = t + syncsize 
			b = b + syncsize
			self.Main.OutPosition = 0		
		end
	else
		--only the main window is showing.  do we need to move it?
		if self.Main.OutPosition ~= 0 then
			if self.Main.OutPosition == 1 then
				--shift it down
				t = t + syncsize 
				b = b + syncsize 
			else
				--shift it up
				t = t - syncsize 
				b = b - syncsize 
			end
			self.Main.OutPosition = 0
		end
	end
	
	b = t + mainBase + challengeList + syncDock + padding
	self.Main.Window:SetAnchorOffsets(l,t,r,b)
	
	self.Main.Window:SetSizingMinimum(280,b-t)
	self.Main.Window:SetSizingMaximum(3000,b-t)
end

-----------------------------------------------------------------------------------------------
--Zone Map Stuff
-----------------------------------------------------------------------------------------------
function KiruChallenge:OnZoneMapChange()
	local mapAddon = self:GetZoneMapAddon()
	
	if mapAddon  ~= nil then
	
		if mapAddon.arAllowedTypesScaled == nil then return end
		if mapAddon.arAllowedTypesPanning == nil then return end
		if mapAddon.arAllowedTypesSuperPanning == nil then return end
		
		self.ZoneMap.Addon = mapAddon
		if self.ZoneMap.Overlay == nil then
			self.ZoneMap.Map = mapAddon.wndZoneMap
			self.ZoneMap.Overlay = mapAddon.wndZoneMap:CreateOverlayType()
			table.insert(mapAddon.arAllowedTypesScaled, self.ZoneMap.Overlay)
			table.insert(mapAddon.arAllowedTypesPanning, self.ZoneMap.Overlay)
			table.insert(mapAddon.arAllowedTypesSuperPanning, self.ZoneMap.Overlay)
			--new stuff for Drop 4
			local markerCategory = "KiruChallenge"
			mapAddon.tPOITypes[self.ZoneMap.Overlay] = {strSprite = "", eCategory = markerCategory, strType = ""}
			mapAddon.tButtonChecks[markerCategory] = true
			--find the categories dropdown list and its parent container
			local parentCategoryList = mapAddon.wndMain:FindChild("MarkerPaneButtonList")
			local parentCategoryContainer = mapAddon.wndMain:FindChild("MarkersList")
			--extend it by 30px to make room for new button without scrolling
			local l,b,r,t = parentCategoryContainer:GetAnchorOffsets()
			parentCategoryContainer:SetAnchorOffsets(l,b,r,t + 30)
			--add the category to the dropdown
			local wndCurr = mapAddon:FactoryProduce(parentCategoryList, "MarkerBtn", markerCategory)
			wndCurr:SetCheck(true)
			--change the button events to a custom function
			wndCurr:RemoveEventHandler("ButtonCheck",mapAddon)
			wndCurr:RemoveEventHandler("ButtonUncheck",mapAddon)
			wndCurr:AddEventHandler("ButtonCheck","OnKCMarkerBtnCheck",mapAddon)
			wndCurr:AddEventHandler("ButtonUncheck","OnKCMarkerBtnUncheck",mapAddon)
			--decorate the button
			wndCurr:FindChild("MarkerBtnLabel"):SetText("KiruChallenge")
			wndCurr:FindChild("MarkerBtnImage"):SetSprite("CRB_CharacterCreateSprites:sprCC_RaceAurin")
			wndCurr:FindChild("MarkerBtnLabel"):SetTextColor(ApolloColor.new("UI_BtnTextGoldListNormal"))
			--arrange the dropdown list
			parentCategoryList:ArrangeChildrenVert(0)
			--make sure the custom functions are there
			mapAddon.OnKCMarkerBtnCheck = function(zonemap, wndHandler, wndControl, eMouseButton)
				self.ZoneMap.Addon:SetTypeVisibility(self.ZoneMap.Overlay, true)
			end
			mapAddon.OnKCMarkerBtnUncheck = function(zonemap, wndHandler, wndControl, eMouseButton)
				self.ZoneMap.Addon:SetTypeVisibility(self.ZoneMap.Overlay, false)
			end
		end
		self:AddLocalChallengesToMap()
		self:ShowMapChallenges(self.Data.Preferences.ShowMap)
	end
end

--a copy of the zonemap factory, except injecting a Kiru button for different event handling
function KiruChallenge:CreateZoneMapMarkerButton(wndParent, strFormName, tObject)
	local wndNew = wndParent:FindChildByUserData(tObject)
	if not wndNew then
		wndNew = Apollo.LoadForm(self.xmlDoc, strFormName, wndParent, self.ZoneMap.Addon)
		wndNew:SetData(tObject)
	end
	return wndNew
end

function KiruChallenge:OnKiruMarkerBtnCheck( wndHandler, wndControl, eMouseButton )
	--self.ZoneMap.Addon:SetTypeVisibility(self.ZoneMap.Overlay, true)
end

function KiruChallenge:OnKiruMarkerBtnUncheck( wndHandler, wndControl, eMouseButton )
	--self.ZoneMap.Addon:SetTypeVisibility(self.ZoneMap.Overlay, false)
end

function KiruChallenge:GetZoneMapAddon()
	local compatAddons = { "ZoneMap" }
	for idx, addon in pairs(compatAddons) do
		local zoneMap = Apollo.GetAddon(addon)
		if zoneMap ~= nil then return zoneMap end
	end
end

function KiruChallenge:AddLocalChallengesToMap()
	if self.ZoneMap.Map and self.ZoneMap.Overlay then			
		for _, chalListItem in pairs(self.Main.ItemList.Items) do
			self:AddChallengeMap(self.Data.AllChallenges[chalListItem.ChallengeID])
		end
		--add the locked challenges
		if self.DataPackage:GetCount() == 0 then return end
		local locked = self.DataPackage:GetList()
		local map = GameLib.GetCurrentZoneMap()
		local currentZone = map.strName
		for chalID, chal in pairs(locked) do
			if currentZone == chal.ZoneName and chal.ZoneName ~= "Housing" then
				self:AddLockedChallengeMap(chalID, chal)
			end
		end
	end
end

function KiruChallenge:AddChallengeMap(challenge)
	if self.ZoneMap.Map == nil or self.ZoneMap.Overlay == nil then return end
	if self.ZoneMap.Items[challenge.ID] == nil then
		local loc = challenge.MapStartLocation or challenge.MapLocation
		if loc ~= nil then
			local tInfo = {
				strIcon = spriteChallenge,
				strIconEdge = "",
				crObject = CColor.new(1, 1, 1, 1),				
				crEdge = CColor.new(1, 1, 1, 1)
			}					
			self.ZoneMap.Items[challenge.ID] = self.ZoneMap.Map:AddObject(self.ZoneMap.Overlay, loc, "Challenge: "..challenge.Name, tInfo, {bNeverShowOnEdge = true, bFixedSizeMedium = true})
		end
	end
end

function KiruChallenge:AddLockedChallengeMap(challengeID, lockedChallenge)
	if self.ZoneMap.Map == nil or self.ZoneMap.Overlay == nil then return end
	if self.ZoneMap.Items[challengeID] == nil then
		local loc = Vector3.New(lockedChallenge.x, 0, lockedChallenge.z)
		if loc ~= nil then
			local tInfo = {
				strIcon = spriteLockedChallenge,
				strIconEdge = "",
				crObject = CColor.new(1, 1, 1, 1),				
				crEdge = CColor.new(1, 1, 1, 1)
			}					
			self.ZoneMap.Items[challengeID] = self.ZoneMap.Map:AddObject(self.ZoneMap.Overlay, loc, "New Challenge: "..lockedChallenge.Name, tInfo, {bNeverShowOnEdge = true, bFixedSizeMedium = true})
		end
	end
end

function KiruChallenge:RemoveChallengeMap(challengeID)
	if self.ZoneMap.Map == nil or self.ZoneMap.Overlay == nil then return end
	if self.ZoneMap.Items[challengeID] == nil then return end
	self.ZoneMap.Map:RemoveObject(self.ZoneMap.Items[challengeID])
	self.ZoneMap.Items[challengeID] = nil
end

function KiruChallenge:ShowMapChallenges(show)
	if self.ZoneMap.Map == nil or self.ZoneMap.Overlay == nil then return end
	local addon = self:GetZoneMapAddon()
	if addon ~= nil then addon:SetTypeVisibility(self.ZoneMap.Overlay, show) end
end

function KiruChallenge:KeybindNub_Clicked( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation )
	if self.KeybindChallenge ~= nil then
		self:ProcessSelected(self.KeybindChallenge)
	end
end

function KiruChallenge:Button_ShowArrow( wndHandler, wndControl, eMouseButton )
    -- make sure the wndControl is valid
    if wndHandler ~= wndControl then
        return
    end
	local challengeID = wndControl:GetData()
	if self.Data.AllChallenges[challengeID].IsHouse then
		local unit = GameLib.GetUnitById(self.CurrentHouse[challengeID].UnitID)
		if unit == nil then self.CurrentHouse[challengeID] = nil end
		unit:ShowHintArrow()
	else
		ChallengesLib.ShowHintArrow(challengeID)
	end
end

function KiruChallenge:Button_EditAuto( wndHandler, wndControl, eMouseButton )
	self:CreateAutoStartEditWindow(wndControl:GetData())
end

function KiruChallenge:CreateAutoStartEditWindow(challengeID)
	self:DestroyAutoStartEditWindow()
	
	local wnd = Apollo.LoadForm(self.xmlDoc, "AutoStartEdit", self.Option.AutoStartList.Items[challengeID].Window, self)

	wnd:SetAnchorPoints(1,0,1,0)
	wnd:SetAnchorOffsets(0,0,195,106)
	
	local chalName = wnd:FindChild("ChallengeName")
	chalName:SetText(self.Data.AllChallenges[challengeID].Name)
	
	local distLabel = wnd:FindChild("DistanceLabel")
	local distance = 10
	if self.Data.Preferences.AutoStart[challengeID] ~= nil then
		local tempDist = self.Data.Preferences.AutoStart[challengeID].D
		if tempDist ~= nil then distance = tempDist end
	end
	distLabel:SetText(distance .. "m")

	local slider = wnd:FindChild("Slider_ASDistance")
	slider:SetValue(distance)
					
    self.Option.EditWindow = { Window = wnd, ChallengeID = challengeID, Distance = distance, DistLabel = distLabel }
end

function KiruChallenge:DestroyAutoStartEditWindow()
	if self.Option.EditWindow ~= nil then self.Option.EditWindow.Window:Destroy() end
	self.Option.EditWindow = nil
end

---------------------------------------------------------------------------------------------------
-- AutoStartEdit Functions
---------------------------------------------------------------------------------------------------

function KiruChallenge:Slider_ASDistance( wndHandler, wndControl, fNewValue, fOldValue )
	local newValue = math.floor(fNewValue)
	local oldValue = math.floor(fOldValue)
	if newValue == oldValue  then return end
	self.Option.EditWindow.Distance = newValue
	self.Option.EditWindow.DistLabel:SetText(newValue .. "m")
end

function KiruChallenge:Button_AutoStartEditDone( wndHandler, wndControl, eMouseButton )
	local challengeID = self.Option.EditWindow.ChallengeID
	
	if self.Data.Preferences.AutoStart[challengeID] ~= nil then
		self.Data.Preferences.AutoStart[challengeID].D = self.Option.EditWindow.Distance
	else
		self.Data.Preferences.AutoStart[challengeID] = { O = 0, D = self.Option.EditWindow.Distance }
	end
	self:DestroyAutoStartEditWindow()
end

-----------------------------------------------------------------------------------------------
-- Stupid Carbine and their unannounced API changes
-----------------------------------------------------------------------------------------------
function KiruChallenge:HelperConvertToTime(nInSeconds, bReturnZero)
	if not bReturnZero and (nInSeconds == nil or nInSeconds == 0) then
		return ""
	end
	
	local strResult = ""
	local nHours = math.floor(nInSeconds / 3600)
	local nMins = math.floor(nInSeconds / 60 - (nHours * 60))
	local nSecs = string.format("%02.f", math.floor(nInSeconds - (nHours * 3600) - (nMins * 60)))

	if nHours ~= 0 then
		strResult = nHours .. ":" .. nMins .. ":" .. nSecs
	else
		strResult = nMins .. ":" .. nSecs
	end

	return strResult
end


-- KiruChallenge Instance
-----------------------------------------------------------------------------------------------
local KiruChallengeInst = KiruChallenge:new()
KiruChallengeInst:Init()
